#include "Skeleton.h"
#include <fstream>
#include <stdio.h>

#define N 4
#define Pr 2  /* Pr * Pc is the number of processors in the B&B worker */
#define Pc 2

/* int infty =  -1 ^ (1<<31); */

int infty = 10000;

typedef int edgemat_t [N][N];

typedef struct {
  int from, to;
} touredge;

typedef struct {
  touredge tour[N];
  int length;
} tour;

typedef struct {
  edgemat_t edge;
  int lb;
  tour solution;
} problem;

inline ostream& operator<<(ostream& os,  const problem prb){
#if Output == 1
  int sum =  0;
  for(int i=0; i<N; i++)
    for(int j=0; j<N; j++) sum+= prb.edge[i][j];
  os << "problem with lb " << prb.lb << " and edgesum " << sum << endl<< flush;
  return os;
#endif
}

typedef struct {
  int f, t;      /* the candidate edge  */
  int len;       /* its residual length */
} edge_triple;

typedef struct {
  int r, c;
  int alpha, beta;  /* row and column contributions to inc */
  int inc;          /* the lb increment if this edge is omitted */
} inc_tuple;

int reduce_row (problem *p, int row)
{
  int col;
  int rowmin;
 
  for (rowmin=infty, col=0; col<N; col++) {
    if (p->edge[row][col] < rowmin) rowmin = p->edge[row][col];
  }
  if (rowmin < infty) {
    for (col=0; col<N; col++) {
      if (p->edge[row][col] < infty)
          p->edge[row][col] -= rowmin;
    }
  }
  return rowmin;
}
 
int reduce_col (problem *p, int col)
{
  int row;
  int colmin;
 
  for (colmin=infty, row=0; row<N; row++) {
    if (p->edge[row][col] < colmin) colmin = p->edge[row][col];
  }
  if (colmin < infty) {
    for (row=0; row<N; row++) {
      if (p->edge[row][col] < infty)
          p->edge[row][col] -= colmin;
    }
  }
  return colmin;
}
 
void reduce_matrix (problem *p)
{
  int row, col, deduct;
 
  for (row = 0; row<N; row++) {
    deduct = reduce_row(p, row);
    if (deduct < infty) p->lb += deduct;
  }
  for (col = 0; col<N; col++) {
    deduct = reduce_col(p, col);
    if (deduct < infty) p->lb += deduct;
  }
}                                                                  

edge_triple mkedgetriple (problem *p, int r, int c)
{
  edge_triple triple;

  triple.f = r; triple.t = c; triple.len = p->edge[r][c];
  return triple;
}

inc_tuple mkinctriple (problem *p, int r, int c)
{
  inc_tuple it;

  it.r = r; it.c = c;
  it.alpha = infty; it.beta = infty; 
  it.inc = (p->edge[r][c] == 0) ? -1: -2;
  /* -2 indicates "not a candidate at all", ie residual length isn't zero */
  return it;
}

int one (int any)
{ 
  return 1;
}

inc_tuple setalpha (int r, int c, inc_tuple it, edge_triple et)
{

  if ((it.inc > -2) && (it.alpha > et.len)) it.alpha = et.len;
  return it;
}

inc_tuple setbeta (int r, int c, inc_tuple it, edge_triple et)
{

  if ((it.inc > -2) && (it.beta > et.len)) it.beta = et.len;
  return it;
}

inc_tuple setinc (inc_tuple it)
{
  
  if (it.inc == -2) {
    ;  /* Not an interesting edge at all */
  } else if ((it.alpha == infty) && (it.beta == infty)) {
    it.inc = -1;
  } else if (it.alpha == infty) {
    it.inc = it.beta;
  } else if (it.beta == infty) {
    it.inc = it.alpha;
  } else {
    it.inc = it.alpha + it.beta;
  }
  return it;
}

inc_tuple bestinc (inc_tuple t1, inc_tuple t2)
{
  /* be careful to make this commutative in cases where inc is the same */

  if (t1.inc > t2.inc) {
     return t1; 
  } else if (t1.inc < t2.inc) {
    return t2;  
  } else if (t1.r > t2.r) {
    return t1;  
  } else if (t1.r < t2.r) {
    return t2;
  } else if (t1.c > t2.c) {
    return t1;
  } else if (t1.c < t2.c) {
    return t2;
  } else {
    cout << "Fold fails - identical tuples\n"; /* should never happen */
    return t1;
  }
}

touredge *pick_edge (problem *p){
  touredge *cand;
  inc_tuple best_edge;
  int i;

  DistributedMatrix<edge_triple> Edges(N, N, curry(mkedgetriple)(p), Pr, Pc);
  DistributedMatrix<inc_tuple>   Incs (N, N, curry(mkinctriple)(p),  Pr, Pc); 

  for (i=0; i<N-1; i++) {
    Incs.zipWithIndexInPlace(Edges, & setalpha);
    Edges.rotateRows(& one);}

  Edges.rotateRows(& one);  /* Back to where they started */

  for (i=0; i<N-1; i++) {
    Incs.zipWithIndexInPlace(Edges, & setbeta);
    Edges.rotateCols(& one);}

  Incs.mapInPlace (& setinc);

  best_edge = Incs.fold(& bestinc);

  if (best_edge.inc == -2) {
    if (Debug) cout << sk_myid << ": Pick edge finished returning NULL " << "\n" << flush;
    return NULL;}
  else {
    cand = (touredge *) malloc (sizeof(touredge));
    cand->from = best_edge.r;
    cand->to   = best_edge.c;
    return cand;}
}

void show_tour (problem *cand)
{
  int i;
 
  printf("Tour has length %d\n\n Edges are ", cand->lb);
  for (i=0; i<N; i++) {
    printf("(%d, %d) ",cand->solution.tour[i].from,cand->solution.tour[i].to);
  }
  printf("\n\n");
  cout << flush;
}

void copy_problem (problem *to, problem *from)
{
  int i, r, c;

  for (r=0; r<N; r++) {
    for (c=0; c<N; c++) {
      to->edge[r][c] = from->edge[r][c];
    }
  }
  to->lb = from->lb;

  for (i = 0; i<N; i++) to->solution.tour[i] = from->solution.tour[i];
  to->solution.length = from->solution.length;
}

problem *useit (touredge *e, problem *p)
{
  int i;
  problem *newp;
 
  newp = (problem *) malloc (sizeof(problem));
  copy_problem (newp, p);
  newp->solution.tour[newp->solution.length++] = *e;
  for (i=0; i<N; i++) {
     newp->edge[e->from][i] = infty;
     newp->edge[i][e->to] = infty;
  }
  newp->edge[e->to][e->from] = infty;
  newp->edge[e->from][e->to] = infty;
  reduce_matrix(newp);
  return newp;
}
 

problem *binit (touredge *e, problem *p)
  {
    problem *newp;
    int deduct;

    newp = (problem *) malloc (sizeof(problem));
    copy_problem (newp, p);
    newp->edge[e->from][e->to] = infty;

    deduct  = reduce_row(newp, e->from);
    if (deduct < infty) newp->lb += deduct;
    deduct = reduce_col(newp, e->to);
    if (deduct < infty) newp->lb += deduct;
    return newp;
  }
                                                                                                       
int is_tour (touredge *cand)
/* assumes cand is a list of N genuine edges */
{
  int enters[N], leaves[N], succ[N];
  int i, next;
 
  for (i=0; i<N; i++) {
    succ[cand[i].from] = cand[i].to;
    leaves[i] = 0;
    enters[i] = 0;
  }
 
  for (i=0, next=0; i<N; i++) {
    leaves[next] = 1;
    next = succ[next];
    enters[next] = 1;
  }
 
  for (i=0; i<N; i++) {
    if (enters[i] == 0) return 0;
    if (leaves[i] == 0) return 0;
  }
  return next==0;
}
                                                   

void show_problem (problem *p)
{
  int i , r, c;

  cout << "----------------------\n\n";

  for (r=0; r<N; r++) {
    for (c=0; c<N; c++) {
      if (p->edge[r][c] < infty) cout << p->edge[r][c] << "  ";
      else cout << "  -";
    }
    cout << "\n";
  }
  cout << "\n LB is " << p->lb << "\n";
  cout << "Tour edges are  ";
  for (i=0; i<p->solution.length; i++) {
   cout << p->solution.tour[i].from << "-" << p->solution.tour[i].to;
  }
  cout << "\n\n----------------------\n";
}


static int started = 0;

problem *getproblem(Empty dummy)
{
  int r, c;
  problem *p;

  if (started==0) {
    started = 1;
    ifstream infile("data.txt");

    p = (problem *) malloc (sizeof(problem));
    for (r=0; r<N; r++)
      for (c=0; c<N; c++) {
       infile >> p->edge[r][c];
        if (p->edge[r][c] == -1) p->edge[r][c] = infty; 
      }
    p->lb = 0;
    p->solution.length = 0;
    reduce_matrix (p);
    return p;
  } else { 
    return NULL;
  }
}

bool betterthan(problem x, problem y)
{
  return x.lb < y.lb ;
}

bool isSolution (problem x)
{
  // if ((x.solution.length == N) && (is_tour(x.solution.tour))) {
  //   cout << "isSolution found a solution\n" << flush;
  // } else {
  //   cout << "isSolution reported no solution\n" << flush;
  // }
  return ((x.solution.length == N) && (is_tour(x.solution.tour)));
}

void generatecases (Empty dummy){
  problem *cand, *with, *without;
  touredge *key_edge;

  do{
    if (Debug) cout << sk_myid << " Trying to get a problem\n" << flush;
    cand = get<problem>();
    if (Debug) cout << sk_myid << " Got a problem" << *cand << endl << flush;
    if (cand && (cand->solution.length < N) ) {
        key_edge = pick_edge(cand);
        // cout << "Picked edge  " << key_edge->from << " " << key_edge->to <<" \n" << flush;
        if (key_edge) {
          with     = useit (key_edge, cand);
          put<problem>(*with);
          if (Debug) cout << sk_myid << " put the with problem " << *with << endl << flush;
          without  = binit (key_edge, cand);
          put<problem>(*without);
          if (Debug) cout << sk_myid << " put the without problem "  << *without << endl << flush;
        }
    }
  } while (cand != NULL);
  if (Debug) cout << "Generate cases got a NULL problem, so terminated\n" << flush;
}

void display_results(problem cand)
{
  if (is_tour(cand.solution.tour)) show_tour(&cand);
  else cout << "No tour found\n";
}

int main(int argc, char **argv){

// *************************************************************
// Test: test of branch and bound skeleton
// ************************************************************* 

  try{
    InitSkeletons(argc,argv);

    Initial<problem>             p1(getproblem); 
    Filter<problem, problem>     p2(generatecases,Pr*Pc);
    BranchAndBound<problem>      p3(p2,2,betterthan,isSolution); 
    Final<problem>               p4(& display_results);  
    Pipe                         p5(p1,p3);
    Pipe                         p6(p5,p4);

    p6.start();

    TerminateSkeletons();
  }
  catch(Exception&){cout << "Exception" << endl << flush;}
}
