//-*- c++ -*-
#include <iostream.h>
#include <stdlib.h>
#include "curry.h"
#include "exception.h"
#include "collection.h"
#include "mpi.h"

#ifdef DEBUG
static int Debug = DEBUG;
#else
static int Debug = 0;
#endif

#ifdef DEBUG
#  if DEBUG > 0
#    define dbg(x) x
#  else
#    define dbg(x)
#  endif
#else
#  define dbg(x)
#endif



//************************************************************************************
//*                                                                                  *
//*                       Draft Proposal of Skeleton Library                         *
//*                                                                                  *
//*                (c) by Herbert Kuchen (kuchen@uni-muenster.de),  Nov. 2001        *
//* including several improvements by Joerg Striegnitz <J.Striegnitz@fz-juelich.de>  *
//*                                                                                  *
//************************************************************************************
//
// $Id: Skeleton.h,v 1.4 2002/03/26 12:44:57 kuchen Exp $
//
// changes:
//  1) 25.02.2002: deadlock eliminated in permutePartition (DA+DM), 
//                 permute, allToAll (DA); send replaced by nonblockingSend
//  2) 04.03.2002: fold and scan (DA) no longer require a commutative argument function
//  3) 07.03.2002: all static variables prefixed with sk_ (e.g. sk_myid, sk_numprocs)
//  4) 08.03.2002: dummy argument type Empty inserted in Filter and Initial
//  5) 11.03.2002: skeleton BranchAndBound (including BBcontroller and Heap) added
//  6) 11.03.2002: interface of Farm changed (it now automatically generates the
//                 required copies of the workers)
//  7) 11.03.2002: nullary argument functions of skeletons replaced by functions
//                 with dummy argument type Empty (simplifies currying of these functions)
//  8) 12.03.2002: more arities for Pipe added
//  8) 13.03.2002: Ported to KCC 3.4
//  9) 13.03.2002: dbg macro added -> avoid checks during runtime
// 10) 13.03.2002: Made send/receive functions inline since they are small and get
//                 frequently called!
// 11) 14.03.2002: error in Atomic and Filter removed (exit now correct)
// 12) 19.03.2002: error in Filter and InitSkeletons removed: sk_noprocs properly initialized
// 13) 21.03.2002: Loop changed (separate propagate and feedback argument functions), error corrected
//                 Par skeleton added
// 14) 25.03.2002: Heap (as well as Stack and Queue) moved to collection.h
//                 Exceptions moved to exception.h;
//                 k-parallel depth first search and breadth first search added
// 15) 19.09.2002: Farmer now with asynchronous workers; MPI_Waitany only waits for submitted requests


// **********************************************************
//  global constants
// **********************************************************

static const int sk_undefined = -1;
// the implementation uses the following message tags:
// MY_TAG is used for ordinary messages containing data.
// STOP_TAG is used to stop the following process;
//   such a message is propagated by the receiver to all its successors.
// TERMINATION_TEST_TAG is sent to a process in order to check whether it has
//   terminated. This is the case, if the process did not produce any results
//   between receiving and returning a message with this tag.
//   This kind of message is propagated by each receiving process to its
//   successors. It is required for the implementation of the Loop
//   skeleton. Note that in the presence of the Filter skeleton, a process
//   may produce an arbitrary number of output values for each consumed input.
//   This complicates the implementation compared to that of e.g. P3L.
// Assumptions:
// 1) a process will not receive a stop message 
//    while a termination test is in progress
// 2) a process will not receive a termination test message 
//    while a termination test is in progress
#define MY_TAG 1
#define STOP_TAG 2
#define TERMINATION_TEST_TAG 3

// *******************************************************************
// ProcessorNo
// *******************************************************************

typedef int ProcessorNo;

static ProcessorNo sk_runningProcessorNo = 0;
static ProcessorNo sk_myid = sk_undefined;  
static ProcessorNo sk_myEntrance = sk_undefined;
static ProcessorNo sk_myExit = sk_undefined;  // note: not necessarily the last process of the skeleton
static int sk_noprocs = 1; // number of processes used by atomic skeleton where sk_myid collaborates
static int sk_numprocs;    // total number of processors used
class Process;
static Process* sk_myProcess = NULL;

// ***********************  auxiliary functions  *********************

inline int log2(int n){int i; for(i=0; n>0; i++) n /= 2; return i-1;} // assumption: n>0

inline void throws(const Exception& e){
  cout << sk_myid << ": " << e << endl << flush;
#if (! defined( __KCC	) && ! defined(_CRAYC) )
  if (! Debug) throw e;
#endif	
}

template <class C1, class C2>
inline C1 proj1_2(C1 a, C2 b){return a;}

template <class C1, class C2>
inline C2 proj2_2(C1 a, C2 b){return b;}

// currently not needed:
// template <class C1, class C2, class C3>
// C3 compose(C3 (* f) C2, C2 (* g)(C1), C1 x){return f(g(x));}

// template <class C0, class C1, class C2, class C3>
// C3 compose2(C3 (* f) C2, C2 (* g)(C0,C1), C0 x, C1 y){return f(g(x,y));}

// *********************************************************************
// Empty   (empty dummy type)
// (needed in nullary argument functions of skeletons;  otherwise
// (i.e. with void instead) currying would not work for these functions)
// *********************************************************************
struct Empty{};

template <class C>
inline C extend(C (*f)(), Empty dummy){return (*f)();}

// *******************  sending and receiving messages  **************
template<class C>
inline void receive(ProcessorNo source, C* valref, int size, MPI_Status* stat) {
   if (source == sk_undefined) throws(UndefinedSourceException());
   dbg(cout << sk_myid << " waiting  for message of size " << size << " from " << source << endl << flush);
   MPI_Recv(valref, size, MPI_BYTE, source, MPI_ANY_TAG, MPI_COMM_WORLD, stat); 
   dbg(cout << sk_myid << " received  message of size "  << size << " with tag " << (*stat).MPI_TAG << " from " << source << endl << flush);
}

template<class C>
inline void nonblockingReceive(ProcessorNo source, C* valref, int size, MPI_Request* req){
   if (source == sk_undefined) throws( UndefinedSourceException());
   MPI_Irecv(valref, size, MPI_BYTE, source, MPI_ANY_TAG, MPI_COMM_WORLD, req);}

template<class C>
inline void send(ProcessorNo destination, C* valref, int size){
   if (destination == sk_undefined) throws(UndefinedDestinationException());
   dbg(cout << sk_myid << " sending  message of size " << size << " to " << destination << endl << flush);
   MPI_Send(valref, size, MPI_BYTE, destination, MY_TAG, MPI_COMM_WORLD);}

template<class C>
inline void syncSend(ProcessorNo destination, C* valref, int size){
   if (destination == sk_undefined) throws( UndefinedDestinationException()); 
   dbg(cout << sk_myid << " sending  message of size " << size << " synchronously to " << destination << endl << flush);
   MPI_Ssend(valref, size, MPI_BYTE, destination, MY_TAG, MPI_COMM_WORLD);}

template<class C>
inline void  nonblockingSend(ProcessorNo destination, C* valref, int size, MPI_Request* req){
   if (destination == sk_undefined) throws( UndefinedDestinationException());
   dbg( cout << sk_myid << " sending  message of size " << size << " to " << destination << endl << flush);
   MPI_Isend(valref, size, MPI_BYTE, destination, MY_TAG, MPI_COMM_WORLD, req);}

inline void sendStop(ProcessorNo destination){
   if (destination == sk_undefined) throws( UndefinedDestinationException());
   dbg( cout << sk_myid << " sending  stop message "  << " to " << destination << endl << flush);
   int dummy;
   MPI_Send(&dummy, sizeof(dummy), MPI_BYTE, destination, STOP_TAG, MPI_COMM_WORLD);}

inline void sendTerminationTest(ProcessorNo destination){
   if (destination == sk_undefined) throws( UndefinedDestinationException());
   dbg(cout << sk_myid << " sending termination test message " << " to " << destination << endl << flush);
   int dummy;
   MPI_Send(&dummy, sizeof(dummy), MPI_BYTE, destination, TERMINATION_TEST_TAG, MPI_COMM_WORLD);}
 
template<class C>
inline void sendReceive(ProcessorNo destination, C* valref1, C* valref2, int size){  //  synchronous!
   if (destination == sk_undefined) throws(UndefinedDestinationException());
   MPI_Status stat;
   // original attempt: (causes deadlock)
   // MPI_Sendrecv(valref1, size, MPI_BYTE, destination, MY_TAG,             
   //              valref2, size, MPI_BYTE, sk_myid,  MY_TAG, MPI_COMM_WORLD, &stat); 
   // 2nd attempt: (works)
   if (destination > sk_myid) {syncSend(destination,valref1,size); 
                            receive(destination,valref2,size,&stat);}
   else  {receive(destination,valref2,size,&stat); syncSend(destination,valref1,size);}}
 

// **********************************************************
// classes to enable partial application and currying
// **********************************************************
// moved to  curry.h

// *********************  more auxiliary functions  **************************

// used in rotateRows and rotateCols (for Distributed Matrix (see below))
  template <class F>
  inline static int auxRotateRows(const Fct1<int,int,F>& f, int blocks, int row, int col){
    return (col + f(row)+ blocks) % blocks;}

  template <class F>
  inline static int auxRotateCols(const Fct1<int,int,F>& f, int blocks, int row, int col){
    return (row + f(col) + blocks) % blocks;}

// ***************************************************************************************************
// Distributed Arrays with data parallel skeletons map, fold, zip, gather, scan, permute, broadcast...
// ***************************************************************************************************
class Process; // forward declaration needed in constructor DistributedArray  

template <class E>
class DistributedArray {
  int n;             // total number of elements of DistributedArray
  E* a;              // E a[localsize];
  int localsize;

  // assumption: block division of array
  int localposition; // position of processor in data parallel group of processors
  int first;         // first index in local partition; assuming division mode: block
  int nextfirst;     // first index in next partition

private:
  inline void init(){
    if ((sk_myExit == sk_undefined) || (sk_myEntrance == sk_undefined)) throws( MissingInitializationException());
      // for simplicity: assuming sk_noprocs is power of 2
      localsize = n/sk_noprocs;             // for simplicity assuming: sk_noprocs divides n
      localposition = sk_myid - sk_myEntrance; 
      first = localposition * localsize;  
      nextfirst = first + localsize;     
      a = new E[localsize];}
public:
  template <class F>
  DistributedArray(int size, const Fct1<int,E,F>& f): n(size) { 
    init();
    for(int i = 0; i<localsize; i++) a[i] = f(i+first);}

  DistributedArray(int size, E (*f)(int)): n(size) { 
    init();
    for(int i = 0; i<localsize; i++) a[i] = f(i+first);}

  DistributedArray(int size, E initial): n(size) { 
    init();
    for(int i = 0; i<localsize; i++) a[i] = initial;}

// The following constructor is in fact scatter. Since non-distributed data structures 
// are replicated on every collaborating processor, there is no need for communication here.
  DistributedArray(int size, const E b[]): n(size) {
    init();
    for(int i = 0; i<localsize; i++) a[i] = b[i+first];}

  DistributedArray(int size): n(size)  {  
    init();}

  inline int getFirst() const {return first;}

  inline int getSize() const {return n;}

  inline int getLocalSize() const {return localsize;}

  inline bool isLocal(int i) const {return (i>= first) && (i<nextfirst);}

  inline void setLocal(int i, E v){  // only applicable to locally available elements 
    a[i] = v;} // using local index; assuming 0 <= i < localsize

  inline void set(int i, E v){ // only allowed for locally available elements; using global index
     if (!isLocal(i))throws( NonLocalAccessException());
     a[i-first] = v;}

  inline E get(int i) const { // only allowed for locally available elements; using global index
     if (!isLocal(i))  throws( NonLocalAccessException());
     return a[i-first];}

  inline E getLocal(int i) const {  // only applicable to locally available elements; 
      return a[i];}          // using local index; assuming 0 <= i < localsize

// ******************** map and variants *************************************
  template <class R, class F>
  inline DistributedArray<R> map(const Fct1<E,R,F>& f) const { // requires no communication
    DistributedArray<R> b(n);  
    for(int i = 0; i<localsize; i++) b.setLocal(i, f(a[i]));
    return b;} 

  template <class R>
  inline DistributedArray<R> map(R (*f)(E)) const {return map(curry(f));}

  template <class R, class F>
  inline DistributedArray<R> mapIndex(const Fct2<int,E,R,F>& f) const { // requires no communication
    DistributedArray<R> b(n); 
    for(int i = 0; i<localsize; i++) b.setLocal(i, f(i+first,a[i]));
    return b;} 

  template <class R>
  inline DistributedArray<R> mapIndex(R (*f)(int,E)) const {return mapIndex(curry(f));}

  template <class F>
  inline void mapInPlace(const Fct1<E,E,F>& f){ // requires no communication
    for(int i = 0; i<localsize; i++) a[i] = f(a[i]);}

  inline void mapInPlace(E (*f)(E)){mapInPlace(curry(f));}

  template <class F>
  inline void mapIndexInPlace(const Fct2<int,E,E,F>& f){ // requires no communication
    for(int i = 0; i<localsize; i++) a[i] = f(i+first,a[i]);} 

  inline void mapIndexInPlace(E (*f)(int,E)){mapIndexInPlace(curry(f));}

  template <class F>
  inline void mapPartitionInPlace(const Fct1<E*,void,F>& f){ f(a); } 

  inline void mapPartitionInPlace(void (*f)(E*)){f(a);}

// ******************** zipWith and variants *************************************
  template <class E2, class R, class F>        // requires no communication
  inline DistributedArray<R> zipWith(const DistributedArray<E2>& b, 
                                     const Fct2<E,E2,R,F> f) const { 
    DistributedArray<R> c(n); 
    for(int i=0; i<localsize; i++) c.setLocal(i, f(a[i],b.getLocal(i)));
    return c;} 

  template <class E2, class R>
  inline DistributedArray<R> zipWith(const DistributedArray<E2>& b,R (*f)(E,E2)) const {
    return zipWith(b,curry(f));} 

  template <class E2, class F>                 // requires no communication
  inline void zipWithInPlace(const DistributedArray<E2>& b, const Fct2<E,E2,E,F> f){ 
    for(int i=0; i<localsize; i++) a[i] = f(a[i],b.getLocal(i));}

  template <class E2>
  inline void zipWithInPlace(DistributedArray<E2>& b,E (*f)(E,E2)){
    zipWithInPlace(b,curry(f));}

  template <class E2, class R, class F>      // requires no communication
  inline DistributedArray<R> zipWithIndex(const DistributedArray<E2>& b, 
                                          const Fct3<int,E,E2,R,F>& f) const { 
    DistributedArray<R> c(n); 
    for(int i=0; i<localsize; i++) c.setLocal(i, f(i+first,a[i],b.getLocal(i)));
    return c;} 

  template <class E2, class R>
  inline DistributedArray<R> zipWithIndex(const DistributedArray<E2>& b,
                                          R (*f)(int,E,E2)) const {
    return zipWithIndex(b,curry(f));} 

  template <class E2, class F>               // requires no communication
  inline void zipWithIndexInPlace(const DistributedArray<E2>& b,
                                  const Fct3<int,E,E2,E,F>& f){ 
    for(int i=0; i<localsize; i++) a[i] = f(i+first,a[i],b.getLocal(i));} 

  template <class E2>
  inline void zipWithIndexInPlace(const DistributedArray<E2>& b,E (*f)(int,E,E2)){
    zipWithIndexInPlace(b,curry(f));} 

// *************************  gather  ********************************************
  void gather(E b[]) const { // E b[n]
    int i;
    for (i=0; i<localsize; i++) b[i+first] = a[i];
    int power = 1;
    int neighbor;
    int log2noprocs = log2(sk_noprocs);
    int *indexstack = new int[n];    // can be avoided by massive index calculations
    indexstack[0] = first; 
    int top = 0;
    int oldtop;
    for(i=0; i<log2noprocs; i++){
      neighbor = sk_myEntrance + (localposition ^ power);
      power *= 2;
      oldtop = top;     
      for (int j=0; j<=oldtop; j++){    // important: communication has to be synchronous!
        sendReceive(neighbor,&indexstack[j],&indexstack[++top],sizeof(int));
        sendReceive(neighbor,&b[indexstack[j]],&b[indexstack[top]],sizeof(E)*localsize);}}
    delete [] indexstack;				
	}

// **************************  fold  *********************************************
  template <class F>
  E fold(const Fct2<E,E,E,F>& f) const { // result on every collaborating processor 
    int i;
    // assumption: f is associative 
    // parallel prefix-like algorithm
    // step 1: local fold
    E result = a[0];
    for(i=1; i<localsize; i++) 
      result = f(result, a[i]);
    // step 2: global folding
    int power = 1;
    int neighbor;
    E result2;
    int log2noprocs = log2(sk_noprocs);
    for(i=0; i<log2noprocs; i++){
      neighbor = sk_myEntrance + (localposition ^ power);
      power *= 2;
      sendReceive(neighbor,&result,&result2,sizeof(E)); // important: this has to be synchronous!
      if (sk_myid < neighbor) result = f(result,result2);
      else result = f(result2,result);} 
    return result;}

  inline E fold(E (*f)(E,E)) const { return fold(curry(f));}

// ****************************  scan  ********************************************
  template <class F>
  void scan(const Fct2<E,E,E,F>& f) const { // result on every collaborating processor 
    int i;
    // assumption: f is associative 
    // step 1: local scan
    for(i=1; i<localsize; i++)
      a[i]  = f(a[i-1], a[i]);
    E sum = a[localsize-1];
    // step 2: global scan
    int power = 1;
    int neighbor;
    E neighborSum;
    int log2noprocs = log2(sk_noprocs);
    for(i=0; i<log2noprocs; i++){
      neighbor = sk_myEntrance + (localposition ^ power);
      power *= 2;
      sendReceive(neighbor,&sum,&neighborSum,sizeof(E)); 
      if (sk_myid > neighbor){ 
        for(int j=0; j<localsize; j++) a[j] = f(neighborSum,a[j]);
        sum = f(neighborSum,sum);}
      else sum = f(sum,neighborSum);}}

  inline E scan(E (*f)(E,E)) const {scan(curry(f));}
// **************************  permute  ********************************************
  template <class F>
  void permutePartition(const Fct1<int,int,F>& f){
    int i;
    MPI_Request req;  MPI_Status stat;
    int receiver = sk_myEntrance + f(localposition);
    if ((receiver < sk_myEntrance) || (receiver >= sk_myEntrance+sk_noprocs)) throws( IllegalPartitionException());
    int sender = sk_undefined;
    for (i=0; i< sk_noprocs; i++) // determine sender by computing invers of f
       if (f(i)+sk_myEntrance == sk_myid) 
         if (sender == sk_undefined) sender = sk_myEntrance + i;
         else throws( IllegalPermuteException()); // f is not bijective 
    if  (sender == sk_undefined) throws( IllegalPermuteException()); // f is not bijective 
    if (receiver != sk_myid){ 
      E* b= new E[localsize];
      for (i=0; i<localsize; i++) b[i] = a[i];
      nonblockingSend(receiver,b,sizeof(E)*localsize,&req); // assumption: messages with same sender and receiver
      receive(sender,a,sizeof(E)*localsize,&stat); // don't pass each other
      MPI_Wait(&req,&stat);
      delete[] b;}} 

  inline void permutePartition(int (*f)(int)){ return permutePartition(curry(f));}

  template <class F>
  void permute(const Fct1<int,int,F>& f){ 
    E* b = new E[localsize];
    int* f_inv = new int[n];
    MPI_Status stat; MPI_Request req;
    int receiver;
    int sender;
    int newposition;
    int l,k;
    for (k=0; k<n; k++) f_inv[k] = sk_undefined;
    for (k=0; k<n; k++){ // check, if f is permutation (Exception, if not)
      int dest = f(k);  
      if ((dest < 0) || (dest >= n) || (f_inv[dest] != sk_undefined)){
        throws( IllegalPermuteException());}
      f_inv[dest] = k;}
    for (k=0; k<localsize; k++){ // send and receive local elements (with minimal buffer space)
      newposition = f(first+k);
      receiver = sk_myEntrance + newposition/localsize; 
      if (receiver != sk_myid) nonblockingSend(receiver,&a[k],sizeof(E),&req);
      else b[newposition-first] = a[k];
      for (l=k; l<n; l+=localsize){ 
        sender = sk_myEntrance + l/localsize;
        newposition = f(l);
        if ((newposition  >= first) && (newposition < nextfirst) && (sender != sk_myid))   
          receive(sender,&b[newposition -first],sizeof(E),&stat);}
      if (receiver != sk_myid) MPI_Wait(&req,&stat);}
    for (k=0; k<localsize; k++)
      a[k] = b[k];
    delete[] b; delete[] f_inv;}  

  inline void permute(int (*f)(int)){ return permute(curry(f));}

// **************************  broadcast  ******************************************** 
  void broadcastPartition(int block){  
     if ((block < 0) || (block >= n/localsize)) throws( IllegalPartitionException());
     unsigned int neighbor;
     unsigned int power = 1;
     unsigned int mask = 1073741822; // 2^30-2
     MPI_Status stat;
     int log2noprocs = log2(sk_noprocs);
     for(int i=0; i<log2noprocs; i++){ 
      if ((localposition & mask) == (block & mask)){
        neighbor = sk_myEntrance + (localposition ^ power);
        if ((localposition & power) == (block & power))
          syncSend(neighbor,a,sizeof(E)*localsize);
        else receive(neighbor,a,sizeof(E)*localsize,&stat);}
      power *= 2;
      mask &= ~power;}}

  void broadcast(int index){  
     int i;
     int block = index / localsize;
     int localindex = index % localsize;  
     if ((block < 0) || (block > n/localsize)) throws( IllegalPartitionException());
     unsigned int neighbor;
     unsigned int power = 1;
     unsigned int mask = 1073741822; // 2^30-2
     MPI_Status stat;
     int log2noprocs = log2(sk_noprocs);
     for(i=0; i<log2noprocs; i++){ 
      if ((localposition & mask) == (block & mask)){
        neighbor = sk_myEntrance + (localposition ^ power);
        if ((localposition & power) == (block & power))
          syncSend(neighbor,&a[localindex],sizeof(E));
        else receive(neighbor,&a[localindex],sizeof(E),&stat);}
      power *= 2;
      mask &= ~power;}
     for (i=0; i<localsize; i++) a[i] = a[localindex];} 

// ************************   allToAll  *****************************************
  // allToAll: each collaborating processor sends a block of elements to every other processor;
  //   the beginnings of all blocks are specified by an index array;
  //   the received blocks are concatenated without gaps in arbitrary order.
  void allToAll(const DistributedArray<int *>& Index, E dummy){ 
                // DistributedArray<int [sk_noprocs+1]> Index(sk_noprocs)   
    MPI_Status stat;
    E* b = new E[localsize]; 
    int i, current; 
    int start = (Index.get(localposition))[localposition];
    int end = (Index.get(localposition))[localposition+1]; 
    int no1 = end-start;
    int no2;
    for (current=0; current<no1; current++) b[current] = a[start+current]; // keep own block
    for (i=1; i<sk_noprocs; i++){ // exchange blocks with all others  
      start = (Index.get(localposition))[localposition^i];
      end = (Index.get(localposition))[(localposition^i)+1];
      no1 = end-start;
      sendReceive(sk_myid^i,&no1,&no2,sizeof(int));
      if (current + no2 > localsize) throws(IllegalAllToAllException());
      if (sk_myid > (sk_myid^i)){
        if (no1 > 0) syncSend(sk_myid^i,&a[start],sizeof(E)*no1);
        if (no2 > 0) receive(sk_myid^i,&b[current],sizeof(E)*no2,&stat);}
      else { 
        if (no2 > 0) receive(sk_myid^i,&b[current],sizeof(E)*no2,&stat);
        if (no1 > 0) syncSend(sk_myid^i,&a[start],sizeof(E)*no1);}
      current += no2;}
    for (i=0; i< current; i++) a[i] = b[i];
    for (i=current; i<localsize; i++) a[i] = dummy; // fill gap at the end with dummy
    delete[] b;}

// ******************************  copy  **************************************
  inline DistributedArray<E> copy() const {return new DistributedArray<E>(this);}
   
  inline DistributedArray<E> copyWithGap(int size, E dummy) const {
    DistributedArray<E> C(size,dummy); // gaps are filled with dummy
    int number = (localsize < C.getLocalSize()) ? localsize : C.getLocalSize();
    for (int i=0; i<number; i++) C.setLocal(i,a[i]);
    return C;}    

// **************************  show  ********************************************
  inline void show() const { // alternatively use << (see below)
#if Output == 1
    E* b =  new E[n];
    gather(b); 
    if (sk_myid == sk_myEntrance){
      cout <<"{";
      for(int i=0; i<n-1; i++) cout << b[i] << ",";
      cout  << b[n-1] << "}" << endl << flush;}
    delete[] b;
#endif
  }
}; // *********  end of class DistributedArray *************
  
template <class E>
ostream& operator<<(ostream& os, const DistributedArray<E>& A){
#if Output == 1
  int n = A.getSize();
  E* b =  new E[n];
  A.gather(b); 
  if (sk_myid == sk_myEntrance){
    os << "{";
    for(int i=0; i<n-1; i++) os << b[i] << ",";
    os << b[n-1] << "}" << endl;} 
  delete[] b;
  return os;
#endif
}

// ***************************************************************************************************
// * Distributed Matrix  with data parallel skeletons map, fold, zip, gather, permute, broadcast...  *
// ***************************************************************************************************

template <class E>
class DistributedMatrix {
  int n;                // number of rows
  int m;                // number of columns
  E** a;                // E a[localRows][localCols];
  int localRows;        // number of local rows
  int localCols;        // number of local columns
  int localsize;        //  = localRows*localCols;
  int blocksInRow;      // number of partitions per row
  int blocksInCol;      // number of partitions per column
  
  // assumption: block division of matrix
  int localColPosition; // X position of processor in data parallel group of processors
  int localRowPosition; // Y position of processor in data parallel group of processors
  int localPosition;    // position of processor in data parallel group of processors
  int firstRow;         // first column index in local partition; assuming division mode: block
  int firstCol;         // first row index in local partition; assuming division mode: block
  int nextRow;          // index of first row in next partition
  int nextCol;          // index of first column in next partition

private:
  void init(int rows, int cols){  // initializes distributed matrix (used in constructors)
    if ((sk_myExit == sk_undefined) || (sk_myEntrance == sk_undefined)) throws( MissingInitializationException());
    blocksInRow = cols; blocksInCol = rows;
    // for simplicity: assuming sk_noprocs is a power of 2
    if (cols*rows != sk_noprocs) throws(PartitioningImpossibleException());
    localRows = n / rows;  // for simplicity: assuming rows divides n
    localCols = m / cols;  // for simplicity: assuming cols divides m
    localsize =  localRows*localCols;   
    localColPosition = (sk_myid - sk_myEntrance) % cols; // blocks assigned row by row
    localRowPosition = (sk_myid - sk_myEntrance) / cols;
    localPosition = localRowPosition * cols + localColPosition;
    firstRow = localRowPosition * localRows;
    firstCol = localColPosition * localCols;
    nextRow = firstRow + localRows;
    nextCol = firstCol + localCols;   
    a = new E*[localRows];
    for (int i=0; i<localRows; i++) a[i] = new E[localCols];}

public:  

  template <class F>
  DistributedMatrix(int n0, int m0, const Fct2<int,int,E,F>& f, int rows, int cols): n(n0), m(m0){
    init(rows, cols);
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) a[i][j] = f(i+firstRow,j+firstCol);}

  DistributedMatrix(int n0, int m0, E (*f)(int,int), int rows, int cols): n(n0), m(m0){ 
    init(rows, cols);  
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) a[i][j] = f(i+firstRow,j+firstCol);}

  DistributedMatrix(int n0, int m0, E initial, int rows, int cols): n(n0), m(m0){ 
    init(rows, cols);  
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) a[i][j] = initial;}

// The following constructor is in fact scatter. Since non-distributed data structures 
// are replicated on every collaborating processor, there is no need for communication here.
  DistributedMatrix(int n0, int m0, const E** b, int rows, int cols): n(n0), m(m0){ //E b[n][m]
    init(rows, cols); 
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) a[i][j] = b[i+firstRow][j+firstCol];}


  DistributedMatrix(int n0, int m0, int rows, int cols): n(n0), m(m0){init(rows, cols);}

  inline int getRows() const {return n;}
  
  inline int getCols() const {return m;}

  inline int getFirstRow() const {return firstRow;}

  inline int getFirstCol() const {return firstCol;}

  inline int getLocalRows() const {return localRows;}
 
  inline int getLocalCols() const {return localCols;}

  inline int getBlocksInRow() const { return blocksInRow;}

  inline int getBlocksInCol() const {return blocksInCol;}

  inline void setLocal(int i, int j, E v){  // only applicable to locally available elements 
    a[i][j] = v;} // using local index; assuming 0 <= i < localRows; 0 <= j < localCols

  inline bool isLocal(int i, int j) const {
    return (i>= firstRow) && (i<nextRow) && (j>= firstCol) && (j<nextCol);}

  inline E get(int i, int j) const { // only allowed for locally available elements; using global index
    if (!isLocal(i,j)) throws( NonLocalAccessException());
    return a[i-firstRow][j-firstCol];}

  inline E getLocal(int i, int j) const {  // only applicable to locally available elements; 
    return a[i][j];}     // using local index; assuming 0 <= i < localRows; 0 <= j < localCols

  inline E getLocalGlobal(int i, int j) const {  // j is global index; i is local index ;
    if ((j< firstCol) || (j>=nextCol)) throws( NonLocalAccessException());
    return a[i][j-firstCol];} // assuming 0 <= i < localRows; 

  inline E getGlobalLocal(int i, int j) const {  // i is global index; j is local index ;
    if ((i< firstRow) || (i>=nextRow)) throws( NonLocalAccessException());
    return a[i-firstRow][j];} // assuming 0 <= j < localCols; 

// ******************** map and variants *************************************
  template <class R, class F>
  inline DistributedMatrix<R> map(const Fct1<E,R,F>& f) const { // requires no communication
    DistributedMatrix<R> b(n,m,blocksInRow,blocksInCol);  // causes implicit call of constructor
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) b.setLocal(i,j, f(a[i][j]));
    return b;} 

  template <class R>
  inline DistributedMatrix<R> map(R (*f)(E)) const {return map(curry(f));}

  template <class R, class F>              // requires no communication
  inline DistributedMatrix<R> mapIndex(const Fct3<int,int,E,R,F>& f) const { 
    DistributedMatrix<R> b(n,m,blocksInRow,blocksInCol);  
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) b.setLocal(i,j, f(i+firstRow,j+firstCol,a[i][j]));
    return b;} 

  template <class R>
  inline DistributedMatrix<R> mapIndex(R (*f)(int,int,E)) const {return mapIndex(curry(f));}

  template <class F>
  inline void mapInPlace(const Fct1<E,E,F>& f){ // requires no communication 
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) a[i][j] = f(a[i][j]);} 

  inline void mapInPlace(E (*f)(E)){mapInPlace(curry(f));}

  template <class F>
  inline void mapIndexInPlace(const Fct3<int,int,E,E,F>& f){ // requires no communication 
    for(int i = 0; i<localRows; ++i)
      for(int j = 0; j<localCols; ++j) a[i][j] = f(i+firstRow,j+firstCol,a[i][j]);} 

  inline void mapIndexInPlace(E (*f)(int,int,E)){mapIndexInPlace(curry(f));}

  template <class F>
  inline void mapPartitionInPlace(const Fct1<E**,void,F>& f){ f(a); } 

  inline void mapPartitionInPlace(void (*f)(E**)){f(a);}


// // ******************** zipWith and variants *************************************
  template <class E2, class R, class F>        // requires no communication
  inline DistributedMatrix<R> zipWith(const DistributedMatrix<E2>& b, 
                                          const Fct2<E,E2,R,F>& f) const { 
    DistributedMatrix<R> c(n,m,blocksInRow,blocksInCol); 
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) c.setLocal(i, j, f(a[i][j],b.getLocal(i,j)));
    return c;} 

  template <class E2, class R>
  inline DistributedMatrix<R> zipWith(const DistributedMatrix<E2>& b,
                                      R (*f)(E,E2)) const {
    return zipWith(b,curry(f));} 

  template <class E2, class F>                 // requires no communication
  inline void zipWithInPlace(const DistributedMatrix<E2>& b, 
                             const Fct2<E,E2,E,F>& f){ 
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) a[i][j] = f(a[i][j],b.getLocal(i,j));}

  template <class E2>
  inline void zipWithInPlace(const DistributedMatrix<E2>& b,
                             E (*f)(E,E2)){
    zipWithInPlace(b,curry(f));}

  template <class E2, class R, class F>      // requires no communication
  inline DistributedMatrix<R> zipWithIndex(const DistributedMatrix<E2>& b, 
                                               const Fct4<int,int,E,E2,R,F>& f) const { 
    DistributedMatrix<R> c(n,m,blocksInRow,blocksInCol);  
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) 
        c.setLocal(i,j, f(i+firstRow,j+firstCol,a[i][j],b.getLocal(i,j)));
    return c;} 

  template <class E2, class R>
  inline DistributedMatrix<R> zipWithIndex(const DistributedMatrix<E2>& b,
                                           R (*f)(int,int,E,E2)) const {
    return zipWithIndex(b,curry(f));} 

  template <class E2, class F>               // requires no communication
  inline void zipWithIndexInPlace(const DistributedMatrix<E2>& b, 
                                  const Fct4<int,int,E,E2,E,F> f){  
    for(int i = 0; i<localRows; i++)
      for(int j = 0; j<localCols; j++) 
        a[i][j] = f(i+firstRow,j+firstCol,a[i][j],b.getLocal(i,j));}

  template <class E2>
  inline void zipWithIndexInPlace(const DistributedMatrix<E2>& b,
                                  E (*f)(int,int,E,E2)){
    zipWithIndexInPlace(b,curry(f));} 


// *************************  gather  ********************************************
  struct Buffer{
      int row;
      int col;};

  void gather(E** b) const { // E b[n][m]
    int i, j, k,l;
    for(i = 0; i<localRows; i++)
      for(j = 0; j<localCols; j++) b[i+firstRow][j+firstCol] = a[i][j];
    // prepare communication buffers and pointers to their components
    Buffer inbuf;
    Buffer outbuf;
    E* inmatrix  = new E[n*m];
    E* outmatrix = new E[n*m];
    int bufCounter;

    int power = 1;
    ProcessorNo neighbor;
    int log2noprocs = log2(sk_noprocs);
    int* indexStack = new int[2*n]; // can be avoided by massive index calculations
    indexStack[0] = firstRow;  indexStack[1] = firstCol; 
    int top = 1;
    int oldtop;
    for(i=0; i<log2noprocs; i++){
      neighbor = sk_myEntrance + (localPosition ^ power);
      power *= 2;
      oldtop = top;     
      for (j=0; j<=oldtop; j+=2){    //  synchronous!
        // fill output buffer 
        outbuf.row = indexStack[j];
        outbuf.col = indexStack[j+1];
        bufCounter = 0;
        for (k=0; k<localRows; k++)
           for (l=0; l<localCols; l++) 
             outmatrix[bufCounter++] = b[outbuf.row+k][outbuf.col+l];
        sendReceive(neighbor,&outbuf,&inbuf,2*sizeof(int));
	sendReceive(neighbor,outmatrix,inmatrix,localsize*sizeof(E));
        // fetch from input buffer
        indexStack[++top] = inbuf.row;
        indexStack[++top] = inbuf.col;
        bufCounter = 0;
        for (k=0; k<localRows; k++)
           for (l=0; l<localCols; l++) 
             b[inbuf.row+k][inbuf.col+l] = inmatrix[bufCounter++];}}
    delete[] indexStack; delete[] inmatrix; delete[] outmatrix;}


// // **************************  fold  *********************************************
  template <class F>
  E fold(const Fct2<E,E,E,F>& f) const { // result on every collaborating processor 
    int i,j;
    // assumption: f is associative and commutative
    // parallel prefix-like algorithm
    // step 1: local fold
    E result = a[0][0];
    for(i = 0; i<localRows; i++)
      for(j = 0; j<localCols; j++) 
        if ((i != 0) || (j != 0))
          result = f(result, a[i][j]);
    // step 2: global folding
    int power = 1;
    int neighbor;
    E result2;
    int log2noprocs = log2(sk_noprocs);
    for(i=0; i<log2noprocs; i++){
      neighbor = sk_myEntrance + (localPosition ^ power);
      dbg(cout << sk_myid << ": fold: neighbor " << neigbor << endl);
      power *= 2;
      sendReceive(neighbor,&result,&result2,sizeof(E)); // important: this has to be synchronous!
      result = f(result,result2);}
    return result;}

  inline E fold(E (*f)(E,E)) const {return fold(curry(f));}

// ****************************  scan  ********************************************
//   template <class F>
//   void scan(const Fct2<E,E,E,F> f) const {... still  missing ...;}

// **************************  permute  ********************************************
  template <class F1, class F2>
  inline 
  void permutePartition(const Fct2<int,int,int,F1>& newRow, 
                        const Fct2<int,int,int,F2>& newCol){
    int i,j, receiver;
    MPI_Request req;  MPI_Status stat;
    receiver = sk_myEntrance + newRow(localRowPosition,localColPosition) * blocksInRow + 
                               newCol(localRowPosition,localColPosition);
    if ((receiver < sk_myEntrance) || (receiver >= sk_myEntrance+sk_noprocs))
      throws(IllegalPartitionException());
    int sender = sk_undefined;
    for (i=0; i<blocksInRow; i++) // determine sender 
      for (j=0; j<blocksInCol; j++){ 
        if (sk_myEntrance + newRow(i,j) * blocksInRow + newCol(i,j) == sk_myid) 
          if (sender == sk_undefined) sender = sk_myEntrance + i*blocksInRow + j;
          else throws(IllegalPermuteException());} // newRow and newCol don't produce a permutation
    if (sender == sk_undefined)
      throws(IllegalPermuteException()); // newRow and newCol don't produce a permutation 
    if (receiver != sk_myid){
      E* b1 = new E[localRows*localCols];
      E* b2 = new E[localRows*localCols];
      int bufCount = 0;
      for (i=0; i<localRows; ++i)
        for(j=0; j<localCols; ++j)  b1[bufCount++] = a[i][j];
      nonblockingSend(receiver,&b1[0],sizeof(E)*localRows*localCols,&req); // assumption: messages with same sender
      receive(sender,&b2[0],sizeof(E)*localRows*localCols,&stat);          // and receiver don't pass each other
      MPI_Wait(&req,&stat);
      bufCount = 0;
      for (i=0; i<localRows; ++i)
        for(j=0; j<localCols; ++j) a[i][j] = b2[bufCount++];
      delete[] b1; delete[] b2;}}

  inline void permutePartition(int (*f)(int,int), int (*g)(int,int)){
    permutePartition(curry(f),curry(g));}
 
// **************************  rotate  ********************************************
  template <class F>
  inline void rotateRows(const Fct1<int,int,F>& f){
    // Most C++ compilers need an explicit cast here :(
    // I have to check what the standard says ... (JS)
    permutePartition(curry( (int (*)(int,int)) proj1_2<int,int>),
                     curry( (int (*)(const Fct1<int,int, F>&,
		                                 int,int,int)) auxRotateRows<F>)  (f)(blocksInRow)
	            ); 
  }

  inline void rotateRows(int (*f)(int)){rotateRows(curry(f));}
  
  inline void rotateRows(int cols) { 
    rotateRows(curry((int (*)(int,int))proj1_2<int,int>)(cols));
  }

  template <class F>
  inline void rotateCols(const Fct1<int,int,F>& f) {
   permutePartition(curry((int (*) (const Fct1<int,int,F>&,int,int,int)) auxRotateCols<F>)(f)(blocksInCol),
                    curry((int (*)(int,int)) proj2_2<int,int>));}

  inline void rotateCols(int (*f)(int)) {rotateCols(curry(f));} 

  inline void rotateCols(int rows) {
    rotateCols(curry((int (*)(int,int))proj1_2<int,int>)(rows));
  }

// **************************  broadcast  ******************************************** 
  void broadcastPartition(int blockRow, int blockCol){  
     if ((blockRow < 0) || (blockRow >= blocksInCol) ||
         (blockCol < 0) || (blockCol >= blocksInRow)) throws(IllegalPartitionException());
     int block = blockRow * blocksInRow + blockCol;
     unsigned int neighbor;
     unsigned int power = 1;
     unsigned int mask = 1073741822; // 2^30-2
     MPI_Status stat; 
     E* buf = new E[localRows*localCols];
     int bufCount;
     int i,k,l;
     int log2noprocs = log2(sk_noprocs);
     for(i=0; i<log2noprocs; i++){ 
      if ((localPosition & mask) == (block & mask)){
        neighbor = sk_myEntrance + (localPosition ^ power);
        if ((localPosition & power) == (block & power)){
          bufCount = 0;
          for (k=0; k<localRows; k++)
            for(l=0; l<localCols; l++)  buf[bufCount++] = a[k][l];
          syncSend(neighbor,buf,sizeof(E)*localsize);}
        else {receive(neighbor,buf,sizeof(E)*localsize,&stat);
              bufCount = 0;
              for (k=0; k<localRows; k++)
                for(l=0; l<localCols; l++)  a[k][l]=  buf[bufCount++];}}
      power *= 2;
      mask &= ~power;}
     delete[] buf;}

   void broadcast(int row, int col){  
     int block = row/localRows * (m/localCols) + col/localCols;  
     if ((block < 0) || (block >= sk_noprocs)) throws( IllegalPartitionException()); 
     unsigned int i,j, neighbor;
     unsigned int power = 1;
     unsigned int mask = 1073741822; // 2^30-2
     MPI_Status stat;
     if (block == localPosition) a[0][0] = a[row-firstRow][col-firstCol];
     int log2noprocs = log2(sk_noprocs);
     for(i=0; i<log2noprocs; i++){ 
       if ((localPosition & mask) == (block & mask)){
         neighbor = sk_myEntrance + (localPosition ^ power);
         if ((localPosition & power) == (block & power))
 	   syncSend(neighbor,&a[0][0],sizeof(E));
         else receive(neighbor,&a[0][0],sizeof(E),&stat);}
       power *= 2;
       mask &= ~power;}
     for (i=0; i<localRows; i++)
       for (j=0; j<localCols; j++) a[i][j] = a[0][0];} 

   // possible additional skeletons: broadcastRow, broadcastCol

// **************************  show  ********************************************
  inline void show() const { // alternatively use << (see below)
#if Output == 1
    int n = getRows();  int m = getCols(); int i,j;
    E** b = new E*[n];
    for (i=0; i<n; i++) b[i] = new E[m];
    gather(b); 
    if (sk_myid == sk_myEntrance){  
      for(i=0; i<n; i++){
        cout << "(";
        for(j=0; j<m; j++) cout << b[i][j] << " ";
        cout << ")" << endl;}
      cout << flush;}
    for (i=0; i<n; i++) delete[] b[i];      
    delete[] b;
#endif
  }
};  // *********************** end of class DistributedMatrix **********************
  
template <class E, int n, int m>
inline ostream& operator<<(ostream& os,  const DistributedMatrix<E>& a){
#if Output == 1
  E** b = new E*[n];
  for (i=0; i<n; i++) b[i] = new E[m];
  gather(b); 
  if (sk_myid == sk_myEntrance){  
    for(int i=0; i<n; i++){ 
      os << "(";
      for(int j=0; j<m; j++) os << b[i][j] << " ";
      os << ")" << endl;}}
  for (i=0; i<n; i++) delete[] b[i]; 
  delete[] b;
  return os;
#endif
}


// *************  initialization and termination of skeleton computations  *************
double starttime;
int Problemsize;
char *progname;

void InitSkeletons(int argc, char* argv[]){
  // needs to be called before a skeleton is used
  MPI_Init(&argc,&argv);
  progname = argv[0];
  MPI_Comm_size(MPI_COMM_WORLD,&sk_numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD,&sk_myid);
  sk_myEntrance = 0; sk_myExit = 0;
  sk_noprocs = sk_numprocs;
  if (argc < 2) Problemsize = 8;
  else Problemsize = strtol(argv[1],NULL,10);
  MPI_Bcast(&Problemsize,1,MPI_INT,0,MPI_COMM_WORLD); 
  starttime = MPI_Wtime();
}

void TerminateSkeletons(){
  if (sk_myid == 0)    
    cout << progname << " Problemsize = " << Problemsize << ", Processors = " << sk_numprocs << ", Time = "
         << MPI_Wtime()-starttime << endl << flush;
  MPI_Finalize(); 
  sk_runningProcessorNo = 0;}

// *******************************************************************
// Processes (for user: atomic, pipe, farm; internal: farmer)
// *******************************************************************
  
class Process{ // abstract

  protected:
    ProcessorNo in;
    ProcessorNo out;
    ProcessorNo entrance;
    ProcessorNo exit; 
    bool finished;


  public:    
    Process(): in(sk_undefined), out(sk_undefined){};
    virtual ~Process(){};
    virtual void start() = 0;
    virtual void setIn(ProcessorNo n){in = n;}
    virtual void setOut(ProcessorNo n){out = n;}
    virtual Process* copy() = 0;
    ProcessorNo getIn() const {return in;}
    ProcessorNo getOut() const {return out;}
    ProcessorNo getEntrance() const {return entrance;}
    ProcessorNo getExit() const {return exit;}  
    virtual void show() const = 0; 
};

// ***********************  Atomic Process  *************************
template <class I, class O>
class Atomic: public Process{
 
    DFct1<I,O> fct; 
    int noprocs; // assumption: uses contiguous block of processors

  public:
    Atomic(O (* f)(I), int n):  Process(), fct( Fct1<I,O,O (*)(I)>(f)  ),noprocs(n) {
      entrance = sk_runningProcessorNo; exit = entrance;
      sk_runningProcessorNo += n;}    
    
    Atomic(const DFct1<I,O>& f, int n): fct(f),noprocs(n), Process(){ 
      entrance = sk_runningProcessorNo; exit = entrance; 
      sk_runningProcessorNo += n;}        

    void start(){
      dbg( cout << sk_myid << " in Atomic::start()" << endl << flush);
      I value; int i;
      MPI_Status stat;
      bool testing = false;
      finished = ((sk_myid < entrance) || (sk_myid >= entrance+noprocs));     
      if (!finished) {sk_myEntrance = entrance; sk_myExit = exit;}
      if (Debug && finished) {cout << sk_myid << " not concerned; leaving " << endl << flush;}
      if (finished) return;
      sk_noprocs = noprocs;
      while (!finished){  
        if (sk_myid == sk_myEntrance){
          dbg( cout << sk_myid << " waiting for message from " << in <<  endl << flush);
          receive(in,&value,sizeof(I),&stat);
          if (stat.MPI_TAG == TERMINATION_TEST_TAG) {
            testing = true; sendTerminationTest(out);}
          else if (stat.MPI_TAG == STOP_TAG) { // stop everything
            finished = true;
            for (i=sk_myEntrance+1; i< sk_myEntrance+noprocs; i++) sendStop(i);}
          else // broadcast value to all collaborating data parallel processors; simplistic algorithm
            for (i=sk_myEntrance+1; i< sk_myEntrance+noprocs; i++)
	      send(i,&value,sizeof(I));}
        else {
          receive(sk_myEntrance,&value,sizeof(I),&stat);
          if (stat.MPI_TAG == STOP_TAG) finished = true;}      
        if ((!finished) && (!testing)){
          O result = fct(value);
          dbg(cout << sk_myid << ": intermediate result " << result << endl << flush);
          send(out,&result,sizeof(O));} 
        testing = false;}
      sendStop(out);
      dbg(cout << sk_myid << ": Atomic finished " << endl << flush);}
    
    Process* copy(){return new Atomic(fct,noprocs);}

    void show() const {if (sk_myid == 0) cout << sk_myid << " Atomic " << entrance << endl << flush;}
};

// ****************************  Filter   ******************************
template <class I, class O>
class Filter: public Process{

    DFct1<Empty,void> fct;
    // *fct is assumed to use get() and put() (see below) to process
    // a sequence of inputs and produce a sequence of outputs
    int noprocs; // assumption: uses contiguous block of processors

  public:
    Filter(void (* f)(Empty), int n): fct( Fct1<Empty,void,void (*)(Empty) >(f) ), noprocs(n), Process(){
      entrance = sk_runningProcessorNo; exit = entrance;
      sk_runningProcessorNo += n;} 

    Filter(const DFct1<Empty,void>& f, int n): fct(f), noprocs(n), Process(){
      entrance = sk_runningProcessorNo; exit = entrance;
      sk_runningProcessorNo += n;} 

    void start(){
      Empty dummy;
      dbg( cout << sk_myid << " in Filter::start" << endl << flush );
      finished = ((sk_myid < entrance) || (sk_myid >= entrance+noprocs));   
      if (Debug && finished) {cout << sk_myid <<  "not concerned; leaving"<< endl << flush;}
      if (!finished) {sk_myEntrance = entrance; sk_myExit = exit; sk_myProcess = this;}
      if (finished) return;
      sk_noprocs = noprocs;
      fct(dummy);
      dbg( cout << sk_myid << ": Filter finished " << endl << flush );
    }

    Process* copy(){return new Filter(fct,noprocs);}

    void show() const {if (sk_myid == 0) cout << sk_myid << " Filter " << entrance << endl << flush;}
};

template <class I>
I* get(){ // should only be used in the body of the argument function of Filter
  if (sk_myProcess == NULL) throws( IllegalGetException());
  I* valref = new I;
  MPI_Status stat;
  bool finished = false; int i;

  if (sk_myid == sk_myEntrance){ // important: sk_myEntrance == sk_myExit
    dbg({cout << sk_myid << "waiting for message from " << (*sk_myProcess).getIn() << endl << flush;})
    receive((*sk_myProcess).getIn(),valref,sizeof(I),&stat);
    if (stat.MPI_TAG == TERMINATION_TEST_TAG){
      sendTerminationTest((*sk_myProcess).getOut()); 
      return get<I>();}
    if (stat.MPI_TAG == STOP_TAG) { // stop everything
      finished = true;
      for (i=sk_myEntrance+1; i< sk_myEntrance+sk_noprocs; i++) sendStop(i);}
    else // broadcast value to all collaborating data parallel processors; simplistic algorithm
      for (i=sk_myEntrance+1; i< sk_myEntrance+sk_noprocs; i++)	
        send(i,valref,sizeof(I));}
  else {
    receive(sk_myEntrance,valref,sizeof(I),&stat);
    if (stat.MPI_TAG == STOP_TAG) finished = true;}
  if (finished) {
    if (sk_myid == sk_myEntrance) sendStop((*sk_myProcess).getOut()); 
    return NULL;}
  else return valref;}

template <class O>
inline void put(O value){ // should only be used in the body of the argument function of Filter
  if (sk_myProcess == NULL) throws( IllegalPutException());
  MPI_Status stat;
  if (sk_myid == sk_myExit)
    send((*sk_myProcess).getOut(),&value,sizeof(O));}


// ******************  Initial Atomic Process  ********************
template <class O>
class Initial: public Process{

    DFct1<Empty,O*> fct;

  public:
    Initial(O* (* f)(Empty)) : Process(), fct(Fct1<Empty,O*, O* (*)(Empty) >(f)){
      entrance = sk_runningProcessorNo++; exit = entrance;}  
       
    Initial(const DFct1<Empty,O*>& f): Process(),fct(f) {
      entrance = sk_runningProcessorNo++; exit = entrance;}         

    void start(){
      Empty dummy;
      dbg({cout << sk_myid << " in Initial::start" << endl << flush;})
      finished = !(sk_myid == entrance);   
      if (Debug && finished) {cout << sk_myid <<  "not concerned; leaving"<< endl << flush;}
      if (!finished) {sk_myEntrance = entrance; sk_myExit = exit;}
      while (!finished) {  
        O* result = fct(dummy);
        if (result == NULL) {finished = true; sendStop(out);}
        else send(out,result,sizeof(O));}
      dbg(cout << sk_myid << ": Initial finished " << endl << flush;)
    }   

    Process* copy(){return new Initial(fct);}

    void show() const {if (sk_myid == 0) cout << sk_myid << " Initial " << entrance << endl << flush;}
};

// ******************  Final Atomic Process  ********************
template <class I>
class Final:  public Process{

    DFct1<I,void> fct; 

  public:
    Final(void (* f)(I)): Process(),fct(Fct1<I,void,void (*)(I)>(f)) {
      entrance = sk_runningProcessorNo++; exit = entrance;}       

    Final(const DFct1<I,void>& f): Process(),fct(f){
      entrance = sk_runningProcessorNo++; exit = entrance;}       

   void  start(){
      I value;
      MPI_Status stat;
      dbg({cout << sk_myid << " in Final::start" << endl << flush;})
      finished = !(sk_myid == entrance);   
      if (Debug && finished) {cout << sk_myid <<  "not concerned; leaving"<< endl << flush;}
      if (!finished) {sk_myEntrance = entrance; sk_myExit = exit;}
      while (!finished) {
        dbg({cout << sk_myid << "waiting for message from " << in << endl << flush;})
        receive(in,&value,sizeof(I),&stat);
        if (stat.MPI_TAG == TERMINATION_TEST_TAG);
        else if (stat.MPI_TAG == STOP_TAG) finished = true;
        else {
          dbg({cout << sk_myid << ": final process received: " << value << endl << flush;}) 
          fct(value);}}    
      sk_runningProcessorNo = 0;
      dbg(cout << sk_myid << ": Final finished " << endl << flush;)
      MPI_Barrier(MPI_COMM_WORLD);
      }

    Process* copy(){return new Final(fct);}

    void show() const {if (sk_myid == 0) cout << sk_myid << " Final " << entrance << endl << flush;}
};
      
// *********************   Pipe   ***********************      
class Pipe: public Process{
    Process** p;
    int length;

  public:

    Pipe(Process& p1, Process& p2): Process(){
      length = 2;
      p = new Process* [length];
      p1.setOut(p2.getEntrance());
      p2.setIn(p1.getExit());
      entrance = p1.getEntrance();
      exit = p2.getExit();
      p[0] = &p1; p[1] = &p2;}

    Pipe(Process& p1, Process& p2, Process& p3): Process(){
      length = 3;
      p = new Process* [length];
      p1.setOut(p2.getEntrance());
      p2.setIn(p1.getExit());
      p2.setOut(p3.getEntrance()); 
      p3.setIn(p2.getExit());
      entrance = p1.getEntrance();
      exit = p3.getExit();
      p[0] = &p1; p[1] = &p2; p[2] = &p3;}
  
    // analogous constructors for other arities have to be added

    inline void setOut(ProcessorNo n){out=n; (*(p[length-1])).setOut(n);}

    inline void setIn(ProcessorNo n){in=n;  (*(p[0])).setIn(n);}
   
    void start(){  
      for(int i = 0; i<length; i++) 
        (*(p[i])).start();} 

    Process* copy(){
      return new Pipe(*((*(p[0])).copy()), *((*(p[1])).copy()));}
   
    void show() const { 
      if (sk_myid == 0){
        cout << "Pipe " << entrance << " children: " << length << endl;
        for(int i = 0; i<length; i++){
          cout << "child " << i << (*(p[i])).getEntrance() << endl << flush;
      	  (*(p[i])).show();}}}

};
 
// ***********  SyncFarmer (auxiliary internal process used in SyncFarm) *********** 
template<class I, class O> 
class SyncFarmer:  public Process{
    ProcessorNo* workerIn;
    ProcessorNo* workerOut;
    int length;
    bool* idle;

  public:
    SyncFarmer(int l): length(l){
      workerIn  = new ProcessorNo[length];
      workerOut = new ProcessorNo[length];
      entrance = sk_runningProcessorNo++; exit = entrance;}

    SyncFarmer(){}

    void start(){
      I value; 
      O* result = new O[length]; 
      int i;
      MPI_Status stat;
      MPI_Request* req = new MPI_Request[length+2];
      // req[0]: for sending to (some) worker
      // req[1..length]: for receiving from workers
      // req[length+1]: for receiving from predecessor
      int currentWorker = 0;
      bool testing = false;
      bool* active = new bool[length+2];
      bool sendingToWorker = false;
      int sendFinishedFlag;
      int whoIsReady;
      int terminationTestCounter = 0;
      int terminationCounter = 0;

      dbg({cout << sk_myid << " in SyncFarmer::start" << endl << flush;})
      finished = !(sk_myid == entrance); 
      if (Debug && finished) {cout << sk_myid <<  "not concerned; leaving"<< endl << flush;}
      if (finished) return; 
      for(i=0; i<=length+1; i++) active[i] = false;
      while (!finished) {    
        // wait for message from predecessor or some worker
        for(i=0; i<length; i++){  
          if (!active[i+1]) {
            dbg({cout << sk_myid << ": farmer waiting for message from " 
                           << workerIn[i] << endl; flush;})
            nonblockingReceive(workerIn[i],&result[i],sizeof(O),&req[i+1]);
  	    active[i+1] = true;}}
  
        if (!active[length+1]) { 
          dbg({cout << sk_myid << ": farmer waiting for message from " << in << endl; flush;})
          nonblockingReceive(in,&value,sizeof(I),&req[length+1]); 
          active[length+1] = true;} 
        if (sendingToWorker) // wait for result from worker or successful sending to worker
          MPI_Waitany(length+1,&req[0],&whoIsReady,&stat);
        else{ // wait for message from predecessor or worker
          MPI_Waitany(length+1,&req[1],&whoIsReady,&stat); whoIsReady++;}
        // dispatch based on source of message 
	active[whoIsReady] = false;
        if (whoIsReady == 0){      // message successfully passed to worker;
          sendingToWorker = false; // allow new input from predecessor
          nonblockingReceive(in,&value,sizeof(I),&req[length+1]);}
        else if (whoIsReady == length+1) // message from predecessor
          if (stat.MPI_TAG == TERMINATION_TEST_TAG){ 
            if  ((terminationTestCounter > 0) || (terminationCounter > 0)){
              cout << sk_myid << ": error in skeleton library: request for termination test"
                   << " while other test is in progress!"<< endl << flush; }
             terminationTestCounter = length;
             // here: remaining risk of deadlock (for simplicity accepted for the moment)
             for(i=0; i<length; i++) sendTerminationTest(workerOut[i]);}
          else if (stat.MPI_TAG == STOP_TAG) {
	         if ((terminationTestCounter > 0) || (terminationCounter > 0)){
                   cout << sk_myid << ": error in skeleton library: request for termination test"
                        << " while other test is in progress!"<< endl << flush;}
                 terminationCounter = length;  
                 active[whoIsReady] = true; // inhibit further communication with predecessor
                 for(i=0; i<length; i++) sendStop(workerOut[i]);}
          else{  // simplistic (cyclic) load balancing 
            dbg({cout << sk_myid << ": farmer propates message from " << in 
                              << " to " << workerOut[currentWorker] << endl << flush;})
            sendingToWorker = true;
            active[whoIsReady] = true; // inhibit read from predecessor until message has been sent
            nonblockingSend(workerOut[currentWorker],&value,sizeof(I),&req[0]);
            currentWorker = (currentWorker +1) % length;}
        else // message from worker 	
          if (stat.MPI_TAG == TERMINATION_TEST_TAG){
            if (--terminationTestCounter == 0)
	      sendTerminationTest(out);} 
          else if (stat.MPI_TAG == STOP_TAG) { 
	         active[whoIsReady] = true; // inhibit further communication with responsible worker
                 dbg({cout << sk_myid << ": stop message received from " << workerIn[whoIsReady-1]
                           << " waiting for " << terminationCounter << " stop messages" 
                           << endl << flush;})
                 if (--terminationCounter == 0){
                   finished = true;
		   sendStop(out);}
                 dbg({cout << sk_myid << ": waiting for " << terminationCounter
                                  <<  " stop messages" << endl << flush;})}
          else {dbg({cout << sk_myid << ": farmer propates message from " 
                                 << workerIn[whoIsReady-1] << "  to " << out << endl << flush;})
                send(out,&result[whoIsReady-1],sizeof(O));}}
      delete[] req; delete[] result; delete[] active;
      dbg(cout << sk_myid << ": SyncFarmer finished " << endl << flush;)}

   inline void setWorkerIn(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerIn[i] = p[i];}

   inline void setWorkerOut(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerOut[i] = p[i];}  
 
   Process* copy(){return new SyncFarmer(length);}

   inline void show() const {  
     if (sk_myid == 0) cout << "SyncFarmer (internal)" << entrance <<  endl << flush;}
     
};      

// ********************* SyncFarm *********************** 
template<class I, class O> 
class SyncFarm: public Process{
    Process ** p; 
    int i, length;
    SyncFarmer<I,O>* farmer;

  public:
    SyncFarm(Process& worker, int l): length(l), Process() {
      farmer = new SyncFarmer<I,O>(length); 
     //  cout << sk_myid << "SyncFarm: SyncFarmer created"  << endl << flush; 
      entrance = (*farmer).getEntrance();
      exit = (*farmer).getExit();
      p = new Process*[length];
      p[0] = &worker;
      for (i=1; i<length; i++) 
        p[i]  = worker.copy();
      ProcessorNo* pno1 = new ProcessorNo[length];
      ProcessorNo* pno2 = new ProcessorNo[length];
      for(int i=0; i<length; i++){
        pno1[i] = (*(p[i])).getEntrance();
        pno2[i] = (*(p[i])).getExit();  
        (*(p[i])).setIn(entrance);
        (*(p[i])).setOut(exit);}
      (*farmer).setWorkerOut(pno1);      
      (*farmer).setWorkerIn(pno2);}

    inline void setOut(ProcessorNo n){out=n; (*farmer).setOut(n);}

    inline void setIn(ProcessorNo n){in=n;  (*farmer).setIn(n);}
 
    inline void start(){
      (*farmer).start();
      for(int i = 0; i<length; i++) 
        (*(p[i])).start();} 

    Process* copy(){return new SyncFarm(*((*(p[0])).copy()),length);}

    inline void show() const { 
      if (sk_myid == 0){
        cout << sk_myid << " SyncFarm " << entrance << endl << flush;
        for(int i = 0; i<length; i++) 
          (*(p[i])).show();}}
};

// ***********  Farmer (auxiliary internal process used in Farm) *********** 
// load balancing: the first non-busy worker gets the next task
// each worker may produce an arbitrary number of results for each task
template<class I, class O> 
class Farmer:  public Process{
    ProcessorNo* workerIn;
    ProcessorNo* workerOut;
    int length;

  public:
    Farmer(int l): length(l){
      workerIn  = new ProcessorNo[length];
      workerOut = new ProcessorNo[length];
      entrance = sk_runningProcessorNo++; exit = entrance;}

    Farmer(){}

    void start(){
      int idx = 0;                  // index of worker receiving nexttask
      bool* busy = new bool[length];// worker[i] still having to accept a task? (i=0,..,length-1)
      int busyCounter = 0;          // #workers still having to accept atask
      I value;                      // buffer for input from predecessor
      I* problem = new I[length];   // buffers for output to workers
      O* result  = new O[length];   // buffers for input from workers
      int i;
      MPI_Status stat;
      MPI_Request* request = new MPI_Request[2*length+1]; // request for sending/receiving to/from workers and predecessors
      int requestors = 0;                                 // active requests
      int* requestor = new int[2*length+1];               // requestor[i] tells the reason for request[i]; if requestor[i] in
      // [0..length-1]:        for receiving from workers
      // [length..2*length-1]: for sending to workers
      // [2*length]:           for receiving from predecessor
      bool waitingForWork = false; // from predecessor
      int whoIsReady;
      int reason;
      bool* delivered = new bool[length]; // has worker[i] received a termination test or stop message?
      int terminationTestCounter = 0;     // how many termination testmessages from workers are expected
      int stopCounter = 0;                // how many stop messages from workers are expected
      bool blockInput = false; // block further tasks from predecessor during termination test and stopping

      dbg({cout << sk_myid << " in Farmer::start" << endl << flush;})
      finished = !(sk_myid == entrance);
      if (Debug && finished) {cout << sk_myid <<  "not concerned; leaving"<< endl << flush;}
      if (finished) return;
      for(i=0; i<length; i++) {busy[i] = false;
        dbg({cout << sk_myid << ": farmer waiting for message from " << workerIn[i] << endl << flush;})
        nonblockingReceive(workerIn[i],&result[i],sizeof(O),&request[requestors]);
        requestor[requestors++] = i;}
      while (!finished) {
        if ((!waitingForWork) && (busyCounter < length) && !blockInput){
          // wait for message from predecessor
          dbg({cout << sk_myid << ": farmer waiting for message from " << in << endl << flush;})
          nonblockingReceive(in,&value,sizeof(I),&request[requestors]);
          requestor[requestors++] = 2*length;
          dbg({cout << sk_myid << ": farmer continues after work request " << endl << flush;})
          waitingForWork = true;}
        if (terminationTestCounter > 0)
          for(i=0; i<length; i++)
	    if (!delivered[i] && !busy[i]) {sendTerminationTest(workerOut[i]);
              busy[i] = true; delivered[i] = true;}
        if (stopCounter > 0)
          for(i=0; i<length; i++)
	    if (!delivered[i] && !busy[i]) {sendStop(workerOut[i]);
              busy[i] = true; delivered[i] = true;}
        dbg({cout << sk_myid << ": farmer before MPI_Waitany " << endl << flush;})
        MPI_Waitany(requestors,&request[0],&whoIsReady,&stat);
        reason = requestor[whoIsReady];
        request[whoIsReady] = request[--requestors]; 
        requestor[whoIsReady] = requestor[requestors];
        dbg({cout << sk_myid << ": farmer dispatching based on " << reason << endl << flush;})
        if ((reason >= length) && (reason < 2*length)) { //message successfully passed to worker;
          busy[reason-length] = false; busyCounter--;}
        else if (reason < length) {// received message from worker
          if (stat.MPI_TAG == TERMINATION_TEST_TAG){
            dbg({cout << sk_myid << ": Farmer received termination test message from "
                      << workerIn[reason] << endl << flush;})
	    if (--terminationTestCounter == 0) {sendTerminationTest(out); blockInput = false;}}
	  else if (stat.MPI_TAG == STOP_TAG) {
                   dbg({cout << sk_myid << ": Farmer received stop message from "
                             << workerIn[reason] << endl << flush;})
		   if (--stopCounter == 0) {sendStop(out); blockInput = false; finished = true;}}
          else { // propagate result
                dbg({cout << sk_myid << ": Farmer received result from " << workerIn[reason]
                          << ": " << result[reason] << endl;
                     cout << sk_myid << ": Farmer waiting for message from "
                          << workerIn[reason] << endl <<  flush;})
	        send(out,&result[reason],sizeof(O)); // wait until result delivered
               nonblockingReceive(workerIn[reason],&result[reason],sizeof(O),&request[requestors]);
               requestor[requestors++] = reason;}}
        else { // received message from predecessor
          waitingForWork = false;
          if (stat.MPI_TAG == TERMINATION_TEST_TAG) {terminationTestCounter = length;
            for(i=0; i<length; i++)
	      if (!busy[i])  {sendTerminationTest(workerOut[i]); busy[i] = true; delivered[i] = true;}
		  else delivered[i] = false;}
          else if (stat.MPI_TAG == STOP_TAG) {stopCounter = length;
            for(i=0; i<length; i++)
	      if (!busy[i])  {sendStop(workerOut[i]); busy[i] = true; delivered[i] = true;}
	      else delivered[i] = false;}
  	  else { // ordinary message
            for(i=0; i<length; i++)
              if (!busy[(idx+i+1)%length]) { // propagate message
                idx = (idx+i+1)%length; i= length+1;
                problem[idx] = value;
                nonblockingSend(workerOut[idx],&problem[idx],sizeof(I),&request[requestors]);
                requestor[requestors++] = idx+length;
                busy[idx] = true; busyCounter++;}
	    if (i == length) throws(InternalErrorException());}}} // there should always be a non-busy worker here
      delete[] request; delete[] requestor; delete[] result; delete[] problem; delete[] busy; delete[] delivered; 
      dbg(cout << sk_myid << ": Farmer finished " << endl << flush;)}

   inline void setWorkerIn(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerIn[i] = p[i];}

   inline void setWorkerOut(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerOut[i] = p[i];}  
 
   Process* copy(){return new Farmer(length);}

   inline void show() const {  
     if (sk_myid == 0) cout << "Farmer (internal)" << entrance <<  endl << flush;}
     
};      

// ********************* Farm *********************** 
template<class I, class O> 
class Farm: public Process{
    Process ** p; 
    int i, length;
    Farmer<I,O>* farmer;

  public:
    Farm(Process& worker, int l): length(l), Process() {
      farmer = new Farmer<I,O>(length); 
     //  cout << sk_myid << "Farm: Farmer created"  << endl << flush; 
      entrance = (*farmer).getEntrance();
      exit = (*farmer).getExit();
      p = new Process*[length];
      p[0] = &worker;
      for (i=1; i<length; i++) 
        p[i]  = worker.copy();
      ProcessorNo* pno1 = new ProcessorNo[length];
      ProcessorNo* pno2 = new ProcessorNo[length];
      for(int i=0; i<length; i++){
        pno1[i] = (*(p[i])).getEntrance();
        pno2[i] = (*(p[i])).getExit();  
        (*(p[i])).setIn(entrance);
        (*(p[i])).setOut(exit);}
      (*farmer).setWorkerOut(pno1);      
      (*farmer).setWorkerIn(pno2);}

    inline void setOut(ProcessorNo n){out=n; (*farmer).setOut(n);}

    inline void setIn(ProcessorNo n){in=n;  (*farmer).setIn(n);}
 
    inline void start(){
      (*farmer).start();
      for(int i = 0; i<length; i++) 
        (*(p[i])).start();} 

    Process* copy(){return new Farm(*((*(p[0])).copy()),length);}

    inline void show() const { 
      if (sk_myid == 0){
        cout << sk_myid << " Farm " << entrance << endl << flush;
        for(int i = 0; i<length; i++) 
          (*(p[i])).show();}}
};


// ***********  BBcontroller (auxiliary internal process used in BranchAndBound) *********** 
template<class I> 
class BBcontroller:  public Process{
    ProcessorNo* workerIn;
    ProcessorNo* workerOut;
    int length;
    bool* idle;
    Heap<I>* heap;
    DFct1<I,bool> isSolution;

  public:
    BBcontroller(int l,const DFct2<I,I,bool>& lth, const DFct1<I,bool>& isSol): 
                                                               length(l), isSolution(isSol){
      heap = new Heap<I>(lth);
      workerIn  = new ProcessorNo[length];
      workerOut = new ProcessorNo[length];
      entrance = sk_runningProcessorNo++; exit = entrance;}

    BBcontroller(){}

    void start(){
      bool* idle = new bool[length]; // worker[i] idle? (i=0,..,length-1)
      int idleCounter = length;
      MPI_Request* req = new MPI_Request[2*length];
      // req[0..length-1]: for receiving from workers
      // req[length..2*length-1]: for sending to workers
      MPI_Status stat;
      I* result = new I[length];
      I* problem = new I[length];
      I value; 
      int whoIsReady;
      int i, idleIdx;

      dbg({cout << sk_myid << " in BBcontroller::start" << endl << flush;})
      finished = !(sk_myid == entrance); 
      if (Debug && finished) {cout << sk_myid <<  "not concerned; leaving"<< endl << flush;}
      if (finished) return; 
      for(i=0; i<length; i++) {idle[i] = true;}
      while (!finished) {  
        dbg((*heap).show();)
        // wait for message from predecessor 
        dbg({cout << sk_myid << ": BBcontroller waiting for message from " << in << endl; flush;})
        receive(in,&value,sizeof(I),&stat); 
        if (stat.MPI_TAG == TERMINATION_TEST_TAG) 
          sendTerminationTest(out); // note: this message is only accepted, if all workers are idle
        else if (stat.MPI_TAG == STOP_TAG){ 
          for(i=0; i<length; i++) sendStop(workerOut[i]);
          for(i=0; i<length; i++){ // accept stop messages from workers
            receive(workerIn[i],&result[i],sizeof(I),&stat);
            if (stat.MPI_TAG != STOP_TAG) throws(InternalErrorException());}
          sendStop(out);
          finished = true;}
	else{ // ordinary message (search problem)
          (*heap).insert(value);
          while (((! (*heap).isEmpty()) && (! isSolution((*heap).top()))) || (idleCounter != length)) { 
            dbg({(*heap).show();
	        if ((! (*heap).isEmpty()) && (isSolution((*heap).top())))
                  cout << " top is solution" << endl << flush;
                else cout << " top is no solution" << endl << flush;
                cout << "idleCounter = " << idleCounter << endl << flush; })
            dbg(cout << sk_myid << ": BBcontroller: outer loop" << endl << flush;)
            while ((idleCounter > 0) && (! (*heap).isEmpty()) && (! isSolution((*heap).top()))) { 
              for(i=0; i<length; i++) 
                if (idle[i]) {idleIdx = i; i= length;}
              problem[idleIdx] = (*heap).get();
               dbg(cout << sk_myid << ": BBcontroller sending problem " << problem[idleIdx] 
                        << " to " << workerOut[idleIdx] << endl << flush;)
              nonblockingSend(workerOut[idleIdx],&problem[idleIdx],sizeof(I),&req[idleIdx+length]);
              idle[idleIdx] = false; idleCounter--; 
              dbg(cout << sk_myid << ": BBcontroller waiting for message from " 
                       << workerIn[idleIdx] << endl; flush;)
              nonblockingReceive(workerIn[idleIdx],&result[idleIdx],sizeof(I),&req[idleIdx]);}
            MPI_Waitany(2*length,req,&whoIsReady,&stat);
            if (whoIsReady >= length) // message successfully passed to worker;
              sendTerminationTest(workerOut[whoIsReady-length]); // check, if worker finished
            else // received message from worker 
              if (stat.MPI_TAG == TERMINATION_TEST_TAG){
                dbg({cout << sk_myid << ": BBcontroller received termination test message from "
                          << workerIn[whoIsReady] << endl << flush;})
                idle[whoIsReady] = true; idleCounter++;}
              else{ // new problem or solution received
                (*heap).insert(result[whoIsReady]);
                dbg({cout << sk_myid << ": BBcontroller received problem from " << workerIn[whoIsReady]
                          << ": " << result[whoIsReady] << ", Solution? " 
                          << isSolution(result[whoIsReady]) << endl;
                     cout << sk_myid << ": BBcontroller waiting for message from " 
                          << workerIn[whoIsReady] << endl <<  flush;})
                nonblockingReceive(workerIn[whoIsReady],&result[whoIsReady],sizeof(I),&req[whoIsReady]);}}
	  if ((*heap).isEmpty() || (! isSolution((*heap).top()))) throws(NoSolutionException());
          else {value = (*heap).top();
                send(out,&value,sizeof(I));}}} 
      delete[] req; delete[] result; delete[] idle; delete[] problem;
      dbg(cout << sk_myid << ": BBcontroller finished " << endl << flush;)}

   inline void setWorkerIn(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerIn[i] = p[i];}

   inline void setWorkerOut(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerOut[i] = p[i];}  

   DFct1<I,bool> getIsSolution(){ return isSolution;}   

   BBcontroller<I>* copy(){return new BBcontroller(length,(*heap).getLth(),getIsSolution());}

   inline void show() const {  
     if (sk_myid == 0) cout << "BBcontroller (internal)" << entrance <<  endl << flush;}
     
};      

// ********************* BranchAndBound *********************** 
template <class I> 
class BranchAndBound: public Process{
    Process ** p; 
    int length; 
    BBcontroller<I>* controller;

  private:
    void init(Process& worker){ 
      int i;
      entrance = (*controller).getEntrance();
      exit = (*controller).getExit();
      p = new Process*[length];
      p[0] = &worker;
      for (i=1; i<length; i++) 
        p[i] = worker.copy();
      ProcessorNo* pno1 = new ProcessorNo[length];
      ProcessorNo* pno2 = new ProcessorNo[length];
      for(i=0; i<length; i++){
        pno1[i] = (*(p[i])).getEntrance();
        pno2[i] = (*(p[i])).getExit();  
        (*(p[i])).setIn(entrance);
        (*(p[i])).setOut(exit);}
      (*controller).setWorkerOut(pno1);      
      (*controller).setWorkerIn(pno2);}

  public:
    template <class F1, class F2>
    BranchAndBound(Process& worker, int l, const Fct2<I,I,bool,F1>& lth, 
                   const Fct1<I,bool,F2>& isSolution): length(l), Process(){
      controller = new BBcontroller<I>(length,lth,isSolution); 
      init(worker);}

    BranchAndBound(Process& worker, int l, bool (* lth)(I,I), 
                   bool (* isSol)(I)): length(l), Process() {
      controller = new BBcontroller<I>(length,Fct2<I,I,bool,bool (*)(I,I)>(lth),
                                              Fct1<I,bool,bool (*)(I)>(isSol)); 
      init(worker);}

    template <class F>
    BranchAndBound(Process& worker, int l, const Fct2<I,I,bool,F>& lth, 
                   bool (* isSol)(I)): length(l), Process() {
      controller = new BBcontroller<I>(length,lth,Fct1<I,bool,bool (*)(I)>(isSol)); 
      init(worker);}

    template <class F>
    BranchAndBound(Process& worker, int l, bool (* lth)(I,I), 
                   const Fct1<I,bool,F>& isSolution): length(l), Process(){
      controller = new BBcontroller<I>(length,Fct2<I,I,bool,bool (*)(I,I)>(lth),isSolution); 
      init(worker);}

    BranchAndBound(Process& worker, int l, BBcontroller<I>& ctrl): length(l), controller(&ctrl), Process(){
      init(worker);}

    inline void setOut(ProcessorNo n){out=n; (*controller).setOut(n);}

    inline void setIn(ProcessorNo n){in=n;  (*controller).setIn(n);}
 
    inline void start(){
      (*controller).start();
      for(int i = 0; i<length; i++) 
        (*(p[i])).start();} 

    Process* copy(){return new BranchAndBound(*((*p[0]).copy()),length,*((*controller).copy()));}

    inline void show() const { 
      if (sk_myid == 0){
        cout << sk_myid << " BranchAndBound " << entrance << endl << flush;
        for(int i = 0; i<length; i++) 
          (*(p[i])).show();}}
};

// ***********  Searchcontroller (auxiliary internal process used in Search) *********** 
template<class I> 
class Searchcontroller:  public Process{
    ProcessorNo* workerIn;
    ProcessorNo* workerOut;
    int length;
    bool* idle;
    Collection<I>* coll;
    DFct1<I,bool>  isSolution;
    DFct1<I,bool>  moreSolutions;

  public:
    Searchcontroller(int l,Collection<I>& clct, const DFct1<I,bool>& isSol, const DFct1<I,bool>& moreSol): 
                           coll(&clct), length(l), isSolution(isSol), moreSolutions(moreSol){
      workerIn  = new ProcessorNo[length];
      workerOut = new ProcessorNo[length];
      entrance = sk_runningProcessorNo++; exit = entrance;}

    Searchcontroller(){}

    void start(){
      bool* idle = new bool[length]; // worker[i] idle? (i=0,..,length-1)
      int idleCounter = length;
      MPI_Request* req = new MPI_Request[2*length];
      // req[0..length-1]: for receiving from workers
      // req[length..2*length-1]: for sending to workers
      MPI_Status stat;
      I* result = new I[length];
      I* problem = new I[length];
      I value; 
      int whoIsReady;
      int i, idleIdx;

      dbg({cout << sk_myid << " in Searchcontroller::start" << endl << flush;})
      finished = !(sk_myid == entrance); 
      if (Debug && finished) {cout << sk_myid <<  "not concerned; leaving"<< endl << flush;}
      if (finished) return; 
      for(i=0; i<length; i++) {idle[i] = true;}
      while (!finished) {  
        dbg((*coll).show();)
        // wait for message from predecessor 
        dbg({cout << sk_myid << ": Searchcontroller waiting for message from " << in << endl; flush;})
        receive(in,&value,sizeof(I),&stat); 
        if (stat.MPI_TAG == TERMINATION_TEST_TAG) 
          sendTerminationTest(out); // note: this message is only accepted, if all workers are idle
        else if (stat.MPI_TAG == STOP_TAG){ 
          for(i=0; i<length; i++) sendStop(workerOut[i]);
          for(i=0; i<length; i++){ // accept stop messages from workers
            receive(workerIn[i],&result[i],sizeof(I),&stat);
            if (stat.MPI_TAG != STOP_TAG) throws(InternalErrorException());}
          sendStop(out);
          finished = true;}
	else{ // ordinary message (search problem)
          (*coll).insert(value);
          bool searching = true;
          while (((! (*coll).isEmpty()) || (idleCounter != length)) && searching) {
            dbg({(*coll).show();
                 cout << "idleCounter = " << idleCounter << endl << flush;})
            while ((idleCounter > 0) && (! (*coll).isEmpty()) && searching) { 
              for(i=0; i<length; i++) 
                if (idle[i]) {idleIdx = i; i= length;}
              problem[idleIdx] = (*coll).get();
              dbg(cout << sk_myid << ": Searchcontroller sending problem " << problem[idleIdx] 
                       << " to " << workerOut[idleIdx] << endl << flush;)
              nonblockingSend(workerOut[idleIdx],&problem[idleIdx],sizeof(I),&req[idleIdx+length]);
              idle[idleIdx] = false; idleCounter--; 
              dbg(cout << sk_myid << ": Searchcontroller waiting for message from " 
                       << workerIn[idleIdx] << endl; flush;)
              nonblockingReceive(workerIn[idleIdx],&result[idleIdx],sizeof(I),&req[idleIdx]);}
            MPI_Waitany(2*length,req,&whoIsReady,&stat);
            if (whoIsReady >= length) // message successfully passed to worker;
              sendTerminationTest(workerOut[whoIsReady-length]); // check, if worker finished
            else // received message from worker 
              if (stat.MPI_TAG == TERMINATION_TEST_TAG){
                dbg({cout << sk_myid << ": Searchcontroller received termination test message from "
                        << workerIn[whoIsReady] << endl << flush;})
                idle[whoIsReady] = true; idleCounter++;}
              else{ // new problem or solution received
                value = result[whoIsReady];
                if (isSolution(value)){send(out,&value,sizeof(I));
	           	               if (! moreSolutions(value)) searching = false;}
                else (*coll).insert(result[whoIsReady]);
                dbg({cout << sk_myid << ": Searchcontroller received problem from " << workerIn[whoIsReady]
                          << ": " << result[whoIsReady] << ", Solution? " 
                          << isSolution(result[whoIsReady]) << endl;
                     cout << sk_myid << ": Searchcontroller waiting for message from " 
                          << workerIn[whoIsReady] << endl <<  flush;})
                nonblockingReceive(workerIn[whoIsReady],&result[whoIsReady],sizeof(I),&req[whoIsReady]);}}
	  //if (searching) throws(NoSolutionException()); // don't complain, if no solutions
          while (idleCounter < length){ // wait for remaining messages from workers
            MPI_Waitany(2*length,req,&whoIsReady,&stat);
            if (whoIsReady >= length) // message successfully passed to worker;
              sendTerminationTest(workerOut[whoIsReady-length]); // check, if worker finished
            else // received message from worker 
              if (stat.MPI_TAG == TERMINATION_TEST_TAG){
                dbg({cout << sk_myid << ": Searchcontroller received termination test message from "
                        << workerIn[whoIsReady] << endl << flush;})
                idle[whoIsReady] = true; idleCounter++;}
              else // ignore new problem or solution
                nonblockingReceive(workerIn[whoIsReady],&result[whoIsReady],sizeof(I),&req[whoIsReady]);}}} 
      delete[] req; delete[] result; delete[] idle; delete[] problem;
      dbg(cout << sk_myid << ": Searchcontroller finished " << endl << flush;)}

   inline void setWorkerIn(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerIn[i] = p[i];}

   inline void setWorkerOut(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerOut[i] = p[i];}  

   DFct1<I,bool> getIsSolution(){ return isSolution;}   
   DFct1<I,bool> getMoreSolutions(){ return moreSolutions;}   

   Searchcontroller<I>* copy(){
     Collection<I>* colptr = (*coll).fresh();
     return new Searchcontroller(length,*colptr,
                                 getIsSolution(),getMoreSolutions());}

   inline void show() const {  
     if (sk_myid == 0) cout << "Searchcontroller (internal)" << entrance <<  endl << flush;}
     
};      

// ********************* Search *********************** 
template <class I> 
class Search: public Process{
    Process ** p; 
    int length; 
    Searchcontroller<I>* controller;

  private:
    void init(Process& worker){ 
      int i;
      entrance = (*controller).getEntrance();
      exit = (*controller).getExit();
      p = new Process*[length];
      p[0] = &worker;
      for (i=1; i<length; i++) 
        p[i] = worker.copy();
      ProcessorNo* pno1 = new ProcessorNo[length];
      ProcessorNo* pno2 = new ProcessorNo[length];
      for(i=0; i<length; i++){
        pno1[i] = (*(p[i])).getEntrance();
        pno2[i] = (*(p[i])).getExit();  
        (*(p[i])).setIn(entrance);
        (*(p[i])).setOut(exit);}
      (*controller).setWorkerOut(pno1);      
      (*controller).setWorkerIn(pno2);}

  public:
    template <class F1, class F2>
    Search(Process& worker, int l, Collection<I>& coll, const Fct1<I,bool,F1>& isSolution,
           const Fct1<I,bool,F2>& moreSolutions): length(l), Process(){
      controller = new Searchcontroller<I>(length,coll,isSolution,moreSolutions); 
      init(worker);}

    Search(Process& worker, int l, Collection<I>& coll, bool (* isSol)(I),
                            bool (* moreSol)(I)): length(l), Process() {
      controller = new Searchcontroller<I>(length,coll,Fct1<I,bool,bool (*)(I)>(isSol),
                                                       Fct1<I,bool,bool (*)(I)>(moreSol)); 
      init(worker);}

    template <class F>
    Search(Process& worker, int l, Collection<I>& coll, const Fct1<I,bool,F>& isSolution, 
           bool (* moreSol)(I)): length(l), Process() {
      controller = new Searchcontroller<I>(length,coll,isSolution,
                                           Fct1<I,bool,bool (*)(I)>(moreSol)); 
      init(worker);}

    template <class F>
    Search(Process& worker, int l, Collection<I>& coll, bool (* isSol)(I),
                   const Fct1<I,bool,F>& moreSol): length(l), Process(){
      controller = new Searchcontroller<I>(length,coll,Fct1<I,bool,bool (*)(I)>(isSol),moreSol); 
      init(worker);}

    Search(Process& worker, int l, Searchcontroller<I>& ctrl): length(l), controller(&ctrl), Process(){
      init(worker);}

    inline void setOut(ProcessorNo n){out=n; (*controller).setOut(n);}

    inline void setIn(ProcessorNo n){in=n;  (*controller).setIn(n);}
 
    inline void start(){
      (*controller).start();
      for(int i = 0; i<length; i++) 
        (*(p[i])).start();} 

    Process* copy(){return new Search(*((*p[0]).copy()),length,*((*controller).copy()));}

    inline void show() const { 
      if (sk_myid == 0){
        cout << sk_myid << " Search " << entrance << endl << flush;
        for(int i = 0; i<length; i++) 
          (*(p[i])).show();}}
};

// *********  LoopGuard (auxiliary internal process used in Loop) *********     
template<class C> 
class LoopGuard:  public Process{
    ProcessorNo bodyIn;
    ProcessorNo bodyOut;
    DFct1<C,bool> propagate; 
    DFct1<C,bool> feedback; 

  public:
    LoopGuard(const DFct1<C,bool>& pg, const DFct1<C,bool>& f):  propagate(pg), feedback(f){
      entrance = sk_runningProcessorNo++; exit = entrance;}

    void start(){
      C value; C result;
      MPI_Status stat;
      MPI_Request req[2];
      int whoFinished;
      bool nonactive[2];
      bool finishing = false;
      bool finishingFailed = false;

      finished = !(sk_myid == entrance); 
      if (finished) return;
      nonactive[0] = true; nonactive[1] = true;
      while (!finished) {        
        dbg({cout << sk_myid << ":  waiting for messages from " << bodyIn 
                  << " and "<< in << endl << flush;})
        if (nonactive[0]) {nonblockingReceive(bodyIn,&result,sizeof(C),&req[0]); 
                           nonactive[0] = false;}
        if (nonactive[1]) {nonblockingReceive(in,&value,sizeof(C),&req[1]);
                           nonactive[1] = false;}
        MPI_Waitany(2,req,&whoFinished,&stat);
        if (whoFinished == 1){ // message from predecessor
          nonactive[1] = true;
          if (stat.MPI_TAG == TERMINATION_TEST_TAG)
            if (finishing)  
              {cout << "error in skeleton library: termination test while stopping the system" << endl << flush;}
	    else // asumption: no request for termination test, before the previous is finished
              sendTerminationTest(bodyOut);
          else if (stat.MPI_TAG == STOP_TAG){ 
            // assumption: no stop message, while testing for termination
 	    finishing = true;
            sendTerminationTest(bodyOut);}
          else send(bodyOut,&value,sizeof(C));}
        else {// message received from body 
          nonactive[0] = true;
          if (stat.MPI_TAG == TERMINATION_TEST_TAG) 
            if (finishing) 
              if (finishingFailed) {
                finishingFailed = false; sendTerminationTest(bodyOut);}
              else sendStop(bodyOut);
            else  sendTerminationTest(out);
          else if (stat.MPI_TAG == STOP_TAG) {finished = true; sendStop(out);}
          else {
            if (feedback(result)) send(bodyOut,&result,sizeof(C));
            if (propagate(result))  send(out,&result,sizeof(C));
            if (finishing) finishingFailed = true;}}}
      dbg(cout << sk_myid << ": LoopGuard finished " << endl << flush;)}

  inline void setBodyIn(ProcessorNo p){bodyIn = p;}

  inline void setBodyOut(ProcessorNo p){bodyOut = p;}

  LoopGuard<C>* copy(){return new LoopGuard<C>(propagate,feedback);}
   
   inline void show() const {
     if (sk_myid == 0) cout << sk_myid << "LoopGuard (internal) " << entrance << endl << flush;}     
};      

// ************************  Loop  ************************** 
//
// warning: this skeleton is likely to produce deadlocks. Hence, it
// is questionable, whether it should stay in the skeleton library.
// A deadlock may arise, if the number of messages in the body
// exceeds the total buffer space in the body.
//
template<class C> 
class Loop: public Process{
    Process * body; 
    LoopGuard<C>* guard;

  private:
    void init(){ 
      (*guard).setBodyIn((*body).getExit());
      (*guard).setBodyOut((*body).getEntrance());
      entrance = (*guard).getEntrance();
      exit = (*guard).getExit();
      (*body).setIn(entrance);
      (*body).setOut(exit);}

  public:
    Loop(Process& p, bool (* propagate)(C), bool (* feedback)(C)): body(&p), Process() { 
      guard = new LoopGuard<C>(Fct1<C,bool,bool (*)(C)>(propagate),Fct1<C,bool,bool (*)(C)>(feedback));
      init();}

    Loop(Process& p,  const DFct1<C,bool> propagate, const DFct1<C,bool> feedback): body(&p), Process() {
      guard = new LoopGuard<C>(propagate,feedback);  
      init();}

    Loop(Process& p,  const DFct1<C,bool> propagate,bool (* feedback)(C)): body(&p), Process() {
      guard = new LoopGuard<C>(propagate,Fct1<C,bool,bool (*)(C)>(feedback));  
      init();}

    Loop(Process& p, bool (* propagate)(C) , const DFct1<C,bool> feedback): body(&p), Process() {
      guard = new LoopGuard<C>(Fct1<C,bool,bool (*)(C)>(propagate),feedback);  
      init();}
    Loop(Process& p, LoopGuard<C>& lg): body(&p), guard(&lg), Process(){
      init();}

    inline void setOut(ProcessorNo n){out=n; (*guard).setOut(n);}

    inline void setIn(ProcessorNo n){in=n;  (*guard).setIn(n);}
 
    inline void start(){
      (*guard).start(); (*body).start();} 

    Process* copy(){return new Loop(*((*body).copy()),(*((*guard).copy())));}

    inline void show() const {
      if (sk_myid == 0){
        cout << sk_myid << " Loop " << entrance << endl << flush;
        (*body).show();}}
};

// ***********  ParController (auxiliary internal process used in Par) *********** 
template<class I, class O> 
class ParController:  public Process{
    ProcessorNo* workerIn;
    ProcessorNo* workerOut;
    int length;
    bool* idle;

  public:
    ParController(int l): length(l){
      workerIn  = new ProcessorNo[length];
      workerOut = new ProcessorNo[length];
      entrance = sk_runningProcessorNo++; exit = entrance;}

    ParController(){}

    void start(){
      bool* idle = new bool[length]; // worker[i] idle? (i=0,..,length-1)
      int idleCounter = length;
      MPI_Request* req = new MPI_Request[length]; // for receiving from workers
      MPI_Status stat;
      O* result = new O[length];
      I value; 
      int whoIsReady;
      int i;

      dbg({cout << sk_myid << " in Parcontroller::start" << endl << flush;})
      finished = !(sk_myid == entrance); 
      if (Debug && finished) {cout << sk_myid <<  "not concerned; leaving"<< endl << flush;}
      if (finished) return; 
      while (!finished) {  
        // wait for message from predecessor 
        dbg({cout << sk_myid << ": Parcontroller waiting for message from " << in << endl; flush;})
        receive(in,&value,sizeof(I),&stat); 
        if (stat.MPI_TAG == TERMINATION_TEST_TAG) 
          sendTerminationTest(out); // note: this message is only accepted, if all workers are idle
        else if (stat.MPI_TAG == STOP_TAG){ 
          for(i=0; i<length; i++) sendStop(workerOut[i]);
          for(i=0; i<length; i++){ // accept stop messages from workers
            receive(workerIn[i],&result[i],sizeof(O),&stat);
            if (stat.MPI_TAG != STOP_TAG) throws(InternalErrorException());}
          sendStop(out);
          finished = true;}
	else{ // ordinary message 
          for(i=0; i<length; i++){
            idle[i] = true;
            send(workerOut[i],&value,sizeof(I));
            nonblockingReceive(workerIn[i],&result[i],sizeof(O),&req[i]);}
          idleCounter = 0;
          for(i=0; i<length; i++)
             sendTerminationTest(workerOut[i]);
          while (idleCounter<length) {
            dbg(cout << "idleCounter = " << idleCounter << endl << flush;)
            MPI_Waitany(length,req,&whoIsReady,&stat);
            if (stat.MPI_TAG == TERMINATION_TEST_TAG){
              idle[whoIsReady] = true; idleCounter++;}
            else{ // new solution received
              send(out,&result[whoIsReady],sizeof(O));
              nonblockingReceive(workerIn[whoIsReady],&result[whoIsReady],sizeof(O),&req[whoIsReady]);}}}}
      delete[] req; delete[] result; delete[] idle;
      dbg(cout << sk_myid << ": Parcontroller finished " << endl << flush;)}

   inline void setWorkerIn(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerIn[i] = p[i];}

   inline void setWorkerOut(ProcessorNo* p){
     for(int i=0; i<length; i++)
       workerOut[i] = p[i];}  
 
   Process* copy(){return new ParController(length);}

   inline void show() const {  
     if (sk_myid == 0) cout << "ParController (internal)" << entrance <<  endl << flush;}
     
};      

// ********************* Par *********************** 
template<class I, class O> 
class Par: public Process{
    Process ** p; 
    int i, length;
    ParController<I,O>* controller;

  public:
    Par(Process** worker, int l): length(l), p(worker), Process() {
      controller = new ParController<I,O>(length); 
      entrance = (*controller).getEntrance();
      exit = (*controller).getExit();
      ProcessorNo* pno1 = new ProcessorNo[length];
      ProcessorNo* pno2 = new ProcessorNo[length];
      for(int i=0; i<length; i++){
        pno1[i] = (*(p[i])).getEntrance();
        pno2[i] = (*(p[i])).getExit();  
        (*(p[i])).setIn(entrance);
        (*(p[i])).setOut(exit);}
      (*controller).setWorkerOut(pno1);      
      (*controller).setWorkerIn(pno2);}

    inline void setOut(ProcessorNo n){out=n; (*controller).setOut(n);}

    inline void setIn(ProcessorNo n){in=n;  (*controller).setIn(n);}
 
    inline void start(){
      (*controller).start();
      for(int i = 0; i<length; i++) 
        (*(p[i])).start();} 

    Process* copy(){
      Process** pnew = new Process*[length];
      for (int i=0; i<length; i++) 
        pnew[i]  = (*(p[i])).copy();
      return new Par(pnew,length);}

    inline void show() const { 
      if (sk_myid == 0){
        cout << sk_myid << " Par " << entrance << endl << flush;
        for(int i = 0; i<length; i++) 
          (*(p[i])).show();}}
};

